import datetime
from typing import Dict, Union

from tweepipe import settings
from tweepipe.db import db_client


class BaseRelationParser:
    def __init__(self, db_conn: db_client.DBClient = None):
        """
        Parse emerging relations from tweets.

        Args:
            db_conn (db_client.DBClient, optional): Active db connection. Defaults to None.
        """
        self.db_conn = db_conn

        self.relation_types = ["mentions", "hashtags", "retweets", "replies", "quotes"]
        self.fn_extract_map = {
            k: v
            for (k, v) in zip(
                self.relation_types,
                [
                    self.get_mention_relations,
                    self.get_hashtag_relations,
                    self.get_retweet_relations,
                    self.get_reply_relations,
                    self.get_quote_relations,
                ],
            )
        }

    def get_creation_time_stamp(self, ts: Union[dict, str]) -> datetime.datetime:
        """
        Extract creation date of tweet from object.

        Args:
            ts (Union[dict, str]): timestamp from tweet object to parse.

        Raises:
            ValueError: Raised in case the format of the timestamp is not recognized.

        Returns:
            datetime.datetime: Loaded datetime object from the timestamp.
        """
        if isinstance(ts, dict) and "$date" in ts:
            return datetime.datetime.strptime(
                ts.get("$date"), settings.TWITTER_API_V1_STR_FORMAT
            )
        elif isinstance(ts, str):
            return datetime.datetime.strptime(
                ts, settings.TWITTER_API_ACADEMIC_V2_STR_FORMAT
            )
        else:
            raise ValueError("unrecognized timestamp format")

    def parse_relations_from_tweet(self, tweet: dict) -> dict:
        """
        Extract all retweet, quote, mention, and hashtag relations from tweet.

        Args:
            tweet (dict): Tweet to parse relations from.

        Returns:
            dict: Parsed relations (as timestamped events).
        """
        output = {}
        for relation_type in self.relation_types:
            tmp_relations = self.fn_extract_map.get(relation_type)(tweet)
            # self.db_conn.add_to_collection(relation_type, tmp_relations)
            output[relation_type] = tmp_relations
        return output

    def get_hashtag(self, hashtag: dict) -> str:
        pass

    def get_author_id(self, tweet: dict) -> str:
        pass

    def get_tweet_id(self, tweet: dict) -> str:
        pass

    def get_user_screen_name(self, tweet: dict) -> str:
        return None

    def get_mentioned_user_id(self, mention: dict) -> str:
        pass

    def get_mentioned_user_screen_name(self, mention: dict) -> str:
        pass

    def get_hashtag_relation(
        self, tweet: dict, hashtag: dict, created_at: datetime.datetime
    ) -> dict:
        """
        Extract hashtag relations from tweet.

        Args:
            tweet (dict): Tweet to parse.
            hashtag (dict): Hashtag corresponding to relation.
            created_at (datetime.datetime): Tweet creation datetime.

        Returns:
            dict: Extracted tweet-hashtag relation.
        """
        return {
            "user_id": self.get_author_id(tweet),
            "hashtag": self.get_hashtag(hashtag),
            "tid": self.get_tweet_id(tweet),
            "created_at": created_at,
            "user_screen_name": self.get_user_screen_name(tweet),
        }

    def get_hashtag_relations(self, tweet: dict) -> list:
        """
        Extract all hashtag relations from tweet.

        Args:
            tweet (dict): Tweet to be parsed.

        Returns:
            list: Extracted hashtag relations.
        """
        hashtag_relations = []
        created_at = self.get_creation_time_stamp(tweet["created_at"])
        if "entities" in tweet and "hashtags" in tweet["entities"]:
            for hashtag in tweet["entities"]["hashtags"]:
                hashtag_relations.append(
                    self.get_hashtag_relation(tweet, hashtag, created_at)
                )

        return hashtag_relations

    def get_mention_relation(
        self, tweet: dict, mention: dict, created_at: datetime.datetime
    ) -> dict:
        """
        Extract a mention relation from tweet metadata.

        Args:
            tweet (dict): Tweet object.
            mention (dict): Mention metadata field.
            created_at (datetime.datetime): Tweet creation timestamp.

        Returns:
            dict: Extracted mention relations.
        """
        return {
            "user_id": self.get_author_id(tweet),
            "mentionned_user_screen_name": self.get_mentioned_user_screen_name(mention),
            "mentionned_user_id": self.get_mentioned_user_id(mention),
            "tid": self.get_tweet_id(tweet),
            "created_at": created_at,
            "user_screen_name": self.get_user_screen_name(tweet),
        }

    def get_mention_relations(self, tweet: dict, key: str = "mentions") -> list:
        """
        Extract all mention relations.

        Args:
            tweet (dict): Tweet to be parsed.
            key (str, optional): Name of the mention field. Defaults to "mentions".

        Returns:
            list: Extracted mention relations.
        """
        mention_relations = []
        created_at = self.get_creation_time_stamp(tweet["created_at"])
        if "entities" in tweet and key in tweet["entities"]:
            for mention in tweet["entities"][key]:
                mention_relations.append(
                    self.get_mention_relation(tweet, mention, created_at)
                )

        return mention_relations

    def get_retweet_author_id(self, tweet: dict):
        pass

    def get_retweet_user_screen_name(self, tweet: dict):
        pass

    def get_retweet_id(self, retweet: dict):
        pass

    def get_retweet_relation(
        self, tweet: dict, retweet: dict, created_at: datetime.datetime
    ) -> dict:
        """
        Extract the retweet relation between a tweet and its retweet.

        Args:
            tweet (dict): Tweet object.
            retweet (dict): Retweet object.
            created_at (datetime.datetime): Tweet creation date.

        Returns:
            dict: Extracted retweet relation.
        """
        return {
            "user_id": self.get_author_id(tweet),
            "retweeted_user_id": self.get_retweet_author_id(tweet),
            "retweeted_user_screen_name": self.get_retweet_user_screen_name(tweet),
            "retweet_id": self.get_retweet_id(retweet),
            "tid": self.get_tweet_id(tweet),
            "created_at": created_at,
            "user_screen_name": self.get_user_screen_name(tweet),
        }

    def get_retweet_relations(self, tweet: dict) -> list:
        pass

    def get_reply_relations(self, tweet: dict) -> list:
        pass

    def get_reply_tweet_id(self, reply: dict) -> str:
        pass

    def get_reply_screen_name(self, tweet: dict) -> list:
        pass

    def get_reply_author_id(self, tweet: dict) -> list:
        pass

    def get_quote_screen_name(self, tweet: dict, quote: dict) -> str:
        pass

    def get_quote_tweet_id(self, quote: dict) -> str:
        pass

    def get_quote_author_id(self, quote: dict) -> str:
        # no author id for quoted users with academic api
        return None

    def get_quote_relation(
        self, tweet: dict, quote: dict, created_at: datetime.datetime
    ) -> dict:
        """
        Extract quote relations within a tweet.

        Args:
            tweet (dict): Tweet object.
            quote (dict): Quote metadata fields.
            created_at (datetime.datetime): Tweet creation date.

        Returns:
            dict: Extracted quote relations.
        """
        return {
            "user_id": self.get_author_id(tweet),
            "quoted_tweet_id": self.get_quote_tweet_id(quote),
            "tid": self.get_tweet_id(tweet),
            "created_at": created_at,
            "user_screen_name": self.get_user_screen_name(tweet),
            "quoted_user_id": self.get_quote_author_id(quote),
            "quoted_user_screen_name": self.get_quote_screen_name(tweet, quote),
        }

    def get_reply_relation(
        self, tweet: dict, reply: dict, created_at: datetime.datetime
    ) -> dict:
        """
        Extract reply relations.

        Args:
            tweet (dict): Tweet object.
            reply (dict): Reply metadata.
            created_at (datetime.datetime): Tweet creation date.

        Returns:
            dict: Extract reply relation.
        """
        return {
            "in_reply_to_user_id": self.get_reply_author_id(tweet),
            "in_reply_to_screen_name": self.get_reply_screen_name(tweet),
            "created_at": created_at,
            "user_id": self.get_author_id(tweet),
            "tid": self.get_tweet_id(tweet),
            "user_screen_name": self.get_user_screen_name(tweet),
            "in_reply_to_tweet_id": self.get_reply_tweet_id(reply),
        }

    def get_quote_relations(self, tweet: dict) -> list:
        pass


class AcademicRelationParser(BaseRelationParser):
    def __init__(self, db_conn: db_client.DBClient):
        super().__init__(db_conn)

    def get_tweet_id(self, tweet: dict) -> str:
        return tweet.get("id")

    def get_author_id(self, tweet: dict) -> str:
        return tweet.get("author_id")

    def get_hashtag(self, hashtag: dict) -> str:
        return hashtag.get("tag")

    def get_mentioned_user_id(self, mention: dict) -> str:
        return mention.get("id")

    def get_mentioned_user_screen_name(self, mention: dict) -> str:
        return mention.get("username")

    def get_retweet_id(self, retweet: dict) -> str:
        return retweet.get("id")

    def get_reply_tweet_id(self, reply: dict) -> str:
        return reply.get("id")

    def get_reply_screen_name(self, tweet: dict) -> list:
        try:
            return tweet["text"].split("@")[1].split(" ")[0]
        except IndexError:
            # author is replying to themselves so no mentions, and no username
            return None

    def get_mentions_exist(self, tweet: dict) -> bool:
        return "entities" in tweet and "mentions" in tweet["entities"]

    def get_reply_author_id(self, tweet: dict) -> list:
        try:
            in_reply_to_username = tweet["text"].split("@")[1].split(" ")[0]
            if self.get_mentions_exist(tweet):
                for mention in tweet["entities"]["mentions"]:
                    if mention["username"] == in_reply_to_username:
                        return mention["id"]
            else:
                # no mentions, perhaps the author changed their username
                return None
        except IndexError:
            # author is mentioning themselves, is the author id
            return tweet.get("author_id")

    def get_retweet_author_id(self, tweet: dict) -> str:
        retweeted_username = self.get_retweeted_username_from_text(tweet["text"])
        if self.get_mentions_exist(tweet):
            for mention in tweet["entities"]["mentions"]:
                if mention["username"] == retweeted_username:
                    return mention["id"]
        else:
            # no mentions, perhaps the author changed their username
            return None

    def get_retweeted_username_from_text(self, text: str) -> str:
        # extract the first username retweeted from the text
        return text.split("RT @")[1].split(":")[0]

    def get_retweet_user_screen_name(self, tweet: dict):
        retweeted_username = self.get_retweeted_username_from_text(tweet["text"])
        if self.get_mentions_exist(tweet):
            for mention in tweet["entities"]["mentions"]:
                if mention["username"] == retweeted_username:
                    return mention["username"]
        else:
            # no mentions, perhaps the author changed their username
            return retweeted_username

    def get_reply_relations(self, tweet: dict) -> list:
        reply_relations = []
        if "referenced_tweets" in tweet:
            created_at = super().get_creation_time_stamp(tweet["created_at"])
            for referenced_tweet in tweet["referenced_tweets"]:
                if referenced_tweet["type"] == "replied_to":
                    reply_relations.append(
                        self.get_reply_relation(tweet, referenced_tweet, created_at)
                    )

        return reply_relations

    def get_quote_screen_name(self, tweet: dict, quote: dict) -> str:
        for url in tweet["entities"]["urls"]:
            if quote["id"] in url["expanded_url"]:
                quoted_username = (
                    url["expanded_url"].split("https://twitter.com/")[1].split("/")[0]
                )
                return quoted_username
        return None

    def get_quote_tweet_id(self, quote: dict) -> str:
        return quote.get("id")

    def get_quote_relations(self, tweet: dict) -> list:
        quote_relations = []
        if "referenced_tweets" in tweet:
            created_at = super().get_creation_time_stamp(tweet["created_at"])
            for referenced_tweet in tweet["referenced_tweets"]:
                if referenced_tweet["type"] == "quoted":
                    quote_relations.append(
                        self.get_quote_relation(tweet, referenced_tweet, created_at)
                    )

        return quote_relations

    def get_retweet_relations(self, tweet: dict) -> list:
        retweet_relations = []
        created_at = super().get_creation_time_stamp(tweet["created_at"])
        if "referenced_tweets" in tweet:
            for referenced_tweet in tweet["referenced_tweets"]:
                if referenced_tweet.get("type") == "retweeted":
                    retweet_relations.append(
                        self.get_retweet_relation(tweet, referenced_tweet, created_at)
                    )
        return retweet_relations


class StandardRelationParser(BaseRelationParser):
    def __init__(self, db_conn: db_client.DBClient):
        super().__init__(db_conn)

    def get_tweet_id(self, tweet: dict) -> str:
        return tweet.get("id_str")

    def get_user_screen_name(self, tweet: dict) -> str:
        return tweet.get("user").get("screen_name")

    def get_author_id(self, tweet: dict) -> str:
        return tweet.get("user").get("id_str")

    def get_hashtag(self, hashtag: dict) -> str:
        return hashtag.get("text")

    def get_mentioned_user_id(self, mention: dict) -> str:
        return mention.get("id_str")

    def get_mentioned_user_screen_name(self, mention: dict) -> str:
        return mention.get("screen_name")

    def get_retweet_author_id(self, tweet: dict) -> str:
        return tweet["retweeted_status"]["user"]["id_str"]

    def get_retweet_user_screen_name(self, tweet: dict) -> str:
        return tweet["retweeted_status"]["user"]["screen_name"]

    def get_retweet_id(self, retweet: dict):
        return retweet["id_str"]

    def get_reply_tweet_id(self, reply: dict) -> str:
        return reply["in_reply_to_status_id_str"]

    def get_reply_screen_name(self, tweet: dict) -> list:
        return tweet["in_reply_to_screen_name"]

    def get_reply_author_id(self, tweet: dict) -> list:
        return tweet["in_reply_to_user_id_str"]

    def get_quote_screen_name(self, tweet: dict, quote: dict) -> str:
        return quote["user"]["screen_name"]

    def get_quote_tweet_id(self, quote: dict) -> str:
        return quote["id_str"]

    def get_quote_author_id(self, quote: dict) -> str:
        return quote["user"]["id_str"]

    def get_users_from_tweet(self, tweet: dict) -> list:
        users = list()
        users.append(tweet["user"])
        if "quoted_status" in tweet:
            users.append(tweet["quoted_status"]["user"])

        if "retweeted_status" in tweet:
            users.append(tweet["retweeted_status"]["user"])
            if "quoted_status" in tweet["retweeted_status"]:
                users.append(tweet["retweeted_status"]["quoted_status"]["user"])

        return users

    def get_quotes(self, tweet: dict) -> list:
        quote_relations = []
        if "quoted_status" in tweet:
            created_at = self.get_creation_time_stamp(tweet["created_at"])
            quote_relations.append(
                self.get_quote_relation(tweet, tweet["quoted_status"], created_at)
            )

        return quote_relations

    def get_quote_relations(self, tweet: dict) -> list:
        quote_relations = []
        quote_relations.extend(self.get_quotes(tweet))
        if "retweeted_status" in tweet:
            quote_relations.extend(self.get_quotes(tweet["retweeted_status"]))

        return quote_relations

    def get_replies(self, tweet: dict) -> list:
        reply_relations = []
        if tweet["in_reply_to_status_id"] and tweet["in_reply_to_user_id"]:
            created_at = self.get_creation_time_stamp(tweet["created_at"])
            reply_relations.append(self.get_reply_relation(tweet, tweet, created_at))

        return reply_relations

    def get_reply_relations(self, tweet: dict) -> list:
        reply_relations = []
        reply_relations.extend(self.get_replies(tweet))
        if "quoted_status" in tweet:
            reply_relations.extend(self.get_replies(tweet["quoted_status"]))
        if "retweeted_status" in tweet:
            reply_relations.extend(self.get_replies(tweet["retweeted_status"]))
            if "quoted_status" in tweet["retweeted_status"]:
                reply_relations.extend(
                    self.get_replies(tweet["retweeted_status"]["quoted_status"])
                )

        return reply_relations

    def get_hashtag_relations(self, tweet: dict) -> list:
        hashtag_relations = []
        hashtag_relations.extend(super().get_hashtag_relations(tweet))
        if "quoted_status" in tweet:
            hashtag_relations.extend(
                super().get_hashtag_relations(tweet["quoted_status"])
            )
        if "retweeted_status" in tweet:
            hashtag_relations.extend(
                super().get_hashtag_relations(tweet["retweeted_status"])
            )
            if "quoted_status" in tweet["retweeted_status"]:
                hashtag_relations.extend(
                    super().get_hashtag_relations(
                        tweet["retweeted_status"]["quoted_status"]
                    )
                )

        return hashtag_relations

    def get_mention_relations(self, tweet: dict) -> list:
        """Explore all tweets' structures looking for mentions

        :param tweet: Tweet to process
        :type tweet: dict
        :return: Mention relations found in tweet
        :rtype: list
        """
        mention_relations = []
        mention_relations.extend(
            super().get_mention_relations(tweet, key="user_mentions")
        )
        if "quoted_status" in tweet:
            mention_relations.extend(
                super().get_mention_relations(
                    tweet["quoted_status"], key="user_mentions"
                )
            )
        if "retweeted_status" in tweet:
            mention_relations.extend(
                super().get_mention_relations(
                    tweet["retweeted_status"], key="user_mentions"
                )
            )
            if "quoted_status" in tweet["retweeted_status"]:
                mention_relations.extend(
                    super().get_mention_relations(
                        tweet["retweeted_status"]["quoted_status"], key="user_mentions"
                    )
                )

        return mention_relations

    def get_retweet_relations(self, tweet: dict) -> list:
        """Search for retweet relations in tweet

        :param tweet: Tweet to process
        :type tweet: dict
        :return: retweet relations
        :rtype: list
        """
        retweet_relations = []
        created_at = self.get_creation_time_stamp(tweet["created_at"])
        if "retweeted_status" in tweet:
            retweet = tweet["retweeted_status"]
            retweet_relations.append(
                self.get_retweet_relation(tweet, retweet, created_at)
            )

        return retweet_relations



from datetime import datetime

import re
import joblib
import numpy as np
from typing import Union

from tweepipe import settings
from tweepipe.db import db_client, db_schema
from tweepipe.legacy.botspot.bfreq import _get_bfreq


class BotSpot:
    """Bot Detection Model."""

    def __init__(self, issue="botspot", model_file_path=None):
        """
        Reproduction of the Botometer Lite model.

        Args:
            issue (str, optional): Name of the issue. Defaults to "botspot".
            model_file_path (_type_, optional): Path to mdel. Defaults to None.
        """
        self.issue = issue
        self.db_conn = db_client.DBClient(
            issue=issue,
            schema=db_schema.BOTSPOT_V1,
        )
        self.botspot = joblib.load(model_file_path or settings.BOTSPOT_MODEL)

    def score_user_batch(self, users: Union[list, set]):
        scores = []
        for user in users:
            scores.append(
                {
                    "uid": user.get("id_str"),
                    "username": user.get("screen_name"),
                    "score": float(self.get_user_score(user)),
                }
            )

        self.db_conn.add_scores(scores)

    def get_users_scores(self, users):
        """Users should be represented by full user profiles."""
        return {
            user["id_str"]: float(self.get_user_score(user)) for user in users if user
        }

    def get_user_score(self, user):
        feature_vec = self._get_user_features(user)
        return self.botspot.predict_proba(feature_vec.reshape(-1, len(feature_vec)))[
            0, 1
        ]

    def _get_screen_name_likelihood(self, name):
        bigram_frequencies = [_get_bfreq(name[i : i + 2]) for i in range(len(name) - 1)]
        if len(bigram_frequencies) > 0:
            return np.mean(np.array(bigram_frequencies))
        else:
            return 0.0

    def _get_user_features(self, user, cap_value=1000000):
        feature_vec = np.zeros(20)

        created_at = user["created_at"]
        creation_time = datetime.strptime(created_at, "%a %b %d %H:%M:%S +0000 %Y")
        probe_time = datetime.now()
        age = probe_time - creation_time
        user_age = age.days * 24 + 1

        feature_vec[0] = user["statuses_count"]
        feature_vec[1] = user["followers_count"]
        feature_vec[2] = user["friends_count"]
        feature_vec[3] = user["favourites_count"]
        feature_vec[4] = user["listed_count"]
        feature_vec[5] = 1 if user["default_profile"] else 0
        feature_vec[6] = 1 if user["profile_use_background_image"] else 0
        feature_vec[7] = 1 if user["verified"] else 0
        feature_vec[8] = max(user["statuses_count"] / user_age, cap_value)
        feature_vec[9] = max(user["followers_count"] / user_age, cap_value)
        feature_vec[10] = max(user["friends_count"] / user_age, cap_value)
        feature_vec[11] = max(user["favourites_count"] / user_age, cap_value)
        feature_vec[12] = max(user["listed_count"] / user_age, cap_value)
        feature_vec[13] = user["followers_count"] / max(user["friends_count"], 1)
        feature_vec[14] = len(user["screen_name"])
        feature_vec[15] = len(re.sub("[^0-9]", "", user["screen_name"]))
        feature_vec[16] = len(user["name"])
        feature_vec[17] = len(re.sub("[^0-9]", "", user["name"]))
        feature_vec[18] = len(user["description"])
        feature_vec[19] = self._get_screen_name_likelihood(user["screen_name"])

        return feature_vec



import pymongo


class CollectionPaginator:
    def __init__(
        self,
        collection: pymongo.collection.Collection,
        projection: dict = None,
        sort: list = None,
        batch_size: int = 2048,
    ):
        self.projection = projection
        self.sort = sort
        self.batch_size = batch_size
        self.collection = collection

    def __iter__(
        self,
    ):
        sort = self.sort if self.sort else [("_id", pymongo.ASCENDING)]
        self.cursor = self.collection.find(
            projection=self.projection,
            batch_size=self.batch_size,
            sort=sort,
            allow_disk_use=True,
        )

        return self.cursor

    def __next__(
        self,
    ):
        return next(self.cursor)



from tweepipe.utils import locations
from tweepipe.utils.migration.versions import ApiVersion


class Query:
    def __init__(
        self,
        stream: bool = False,
        search: bool = True,
        api_version: str = ApiVersion.STANDARD_V1a1,
    ):
        """
        _summary_

        Args:
            stream (bool, optional): Whether want to construct a query for stream. Defaults to False.
            search (bool, optional): Whether want to construct a query for search. Defaults to True.
            api_version (str, optional): API version informs how to structure queries. Defaults to ApiVersion.STANDARD_V1a1.
        """
        self.query_args = []
        self.api_version = api_version
        self.stream = stream
        self.search = search

    def add_keywords(self, keywords: list):
        self.query_args.append(" OR ".join(keywords))

    def add_user(self, username: str):
        self.query_args.append(f"from:{username}")

    ######## TWEETS GEOTAGGED QUERY OPERATORS ########

    def add_place(
        self,
    ):
        if (
            self.api_version != ApiVersion.ACADEMIC_V2
            and self.api_version != ApiVersion.ENTERPRISE_POWERTRACK_V2
        ):
            raise ValueError("Cannot use place filter operator.")
        if not self.search:
            raise ValueError("Can only use place filter when searching.")
        return

    def add_point_radius(self, coordinates: tuple, radius: float):
        if (
            self.api_version != ApiVersion.ACADEMIC_V2
            and self.api_version != ApiVersion.ENTERPRISE_POWERTRACK_V2
        ):
            raise ValueError("Cannot use point radius filter operator.")
        if not self.search:
            raise ValueError("Can only use point radius filter when searching.")

        point_radius_str = f"{coordinates[0]},{coordinates[1]},{radius}km"
        self.query_args.append(point_radius_str)

    def add_bounding_box(self, bounding_box: list):
        if (
            self.api_version != ApiVersion.ACADEMIC_V2
            and self.api_version != ApiVersion.ENTERPRISE_POWERTRACK_V2
        ):
            raise ValueError("Cannot use geo filter operator.")
        if not self.search:
            raise ValueError("Can only use geo filter when searching.")
        if len(bounding_box) != 4:
            raise ValueError("Bounding box size must be 4.")

        # filter using bounding boxes
        bbox_str = list(map(str, bounding_box)).join(" ")

        self.query_args.append(f"bounding_box:{bbox_str}")

    def add_tweet_geo_filter(self):
        if (
            self.api_version != ApiVersion.ACADEMIC_V2
            and self.api_version != ApiVersion.ENTERPRISE_POWERTRACK_V2
        ):
            raise ValueError("Cannot use geo filter operator.")
        if not self.search:
            raise ValueError("Can only use geo filter when searching.")

        self.query_args.append("has:geo")

    def add_tweet_geocode(self, coordinates: tuple, radius: float):
        """Add a coordinate and radius to filter geotagged tweets."""

        if self.api_version != ApiVersion.STANDARD_V1a1:
            raise ValueError("Cannot use geocode operator without Standard v1.1 API.")
        if not self.search:
            raise ValueError("Can only use geocode operator when searching tweets.")

        self.query_args.append(f"{coordinates[0]},{coordinates[1]},{radius}km")

    def add_tweet_locations(self, bounding_boxes: list):
        """Add a list of bounding boxes to the chain of query operators."""

        if self.api_version != ApiVersion.STANDARD_V1a1:
            raise ValueError("Cannot use locations operator without Standard v1.1 API.")
        if not self.stream:
            raise ValueError("Need to be streaming data to use this operator.")

        tmp_bbox = ""
        for bbox in bounding_boxes:
            if len(bbox) != 4:
                raise ValueError("Wrong number of coordinates.")
            tmp_bbox_str = list(map(str, bbox)).join(",") + ","
            tmp_bbox += tmp_bbox_str

        self.query_args.append(tmp_bbox)

    def add_place_country(self, country: str):
        if (
            self.api_version != ApiVersion.ENTERPRISE_POWERTRACK_V2
            and self.api_version != ApiVersion.ACADEMIC_V2
        ):
            raise ValueError(
                "Cannot use profile country without Enterprise or Academic API."
            )
        if country not in locations.country_isos:
            raise ValueError("Country ISO code is not recognized.")
        if not self.search:
            raise ValueError("Can only use place country when searching tweets.")

        self.query_args.append(f"place_country:{country}")

    ######## PROFILE GEOLOCATION QUERY OPERATORS ########
    def add_bio_location_keyword(self, keyword: str):
        if self.api_version != ApiVersion.ENTERPRISE_POWERTRACK_V2:
            raise ValueError("Cannot use profile country without Enterprise API.")
        self.query_args.append(f"bio_location:{keyword}")

    def add_bio_keyword(self, keyword: str):
        if self.api_version != ApiVersion.ENTERPRISE_POWERTRACK_V2:
            raise ValueError("Cannot use profile country without Enterprise API.")
        self.query_args.append(f"bio:{keyword}")

    def add_profile_region(self, region: str):
        if self.api_version != ApiVersion.ENTERPRISE_POWERTRACK_V2:
            raise ValueError("Cannot use profile region without Enterprise API.")

        self.query_args.append(f"profile_region:{region}")

    def add_profile_locality(self, locality: str):
        if self.api_version != ApiVersion.ENTERPRISE_POWERTRACK_V2:
            raise ValueError("Cannot use profile country without Enterprise API.")
        pass

    def add_profile_sub_region(self, sub_region: str):
        if self.api_version != ApiVersion.ENTERPRISE_POWERTRACK_V2:
            raise ValueError("Cannot use profile country without Enterprise API.")
        pass

    def add_profile_country(self, country: str):
        if self.api_version != ApiVersion.ENTERPRISE_POWERTRACK_V2:
            raise ValueError("Cannot use profile country without Enterprise API.")
        if country not in locations.country_isos:
            raise ValueError("Country ISO code is not recognized.")

        self.query_args.append(f"profile_country:{country}")

    def add_profile_geo_filter(
        self,
    ):
        if self.api_version != ApiVersion.ENTERPRISE_POWERTRACK_V2:
            raise ValueError("Cannot use profile country without Enterprise API.")

        self.query_args.append(f"has:profile_geo")

    def add_profile_bounding_box(self, bbox: list):
        if self.api_version != ApiVersion.ENTERPRISE_POWERTRACK_V2:
            raise ValueError("Cannot use profile country without Enterprise API.")

        bbox_str = list(map(str, bbox))
        # TODO: check if should join them with spaces or commas
        self.query_args.append(bbox_str.join(","))

    def add_profile_point_radius(self, coordinates: tuple, radius: float):
        """Specify the search coordinates (latitude, longitude) and radius. Note that the radius should
        be specified in kilometers."""
        if self.api_version != ApiVersion.ENTERPRISE_POWERTRACK_V2:
            raise ValueError("Cannot use profile country without Enterprise API.")

        # TODO: check format for the point radius format
        point_radius_str = f"{coordinates[0]},{coordinates[1]},{radius}km"
        self.query_args.append(point_radius_str)

    def get_query_str(self):
        return self.query_args.join(" ")



from tweepipe import settings
from tweepipe.db import db_schema, db_client


def get_follower_retrieval_parallel_kwargs(
    twitter_apis, uids, issue, env_file, include_relations, include_users, cap_count
):
    chunk_size = int(len(uids) / len(twitter_apis))
    kwargs_list = []
    for i in range(len(twitter_apis)):
        kwargs = {
            "twitter_api": twitter_apis[i],
            "include_users": include_users,
            "include_relations": include_relations,
            "cap_count": cap_count,
            "env_file": env_file,
            "issue": issue,
            "uids": uids[i * chunk_size : (i + 1) * chunk_size],
        }
        kwargs_list.append(kwargs)

    return kwargs_list


def fetch_followers(
    twitter_api, uids, include_users, include_relations, cap_count, env_file, issue
):
    if env_file:
        settings.load_config(env_file)

    db_conn = db_client.DBClient(
        issue=issue,
        include_relations=include_relations,
        include_users=include_users,
        schema=db_schema.INDEX_V3,
    )

    for uid in uids:
        pass



from enum import Enum
from typing import Any


class SnPipelineErrorMsg(Enum):
    """Standardise error codes for sutton client operations."""

    UNSUPPORTED_FILE_FORMAT = "File format is not supported."
    UNSUFFICIENT_FREE_RESOURCES = "Less credentials were available than requested."


class SnPipelineError(Exception):
    def __init__(self, message: SnPipelineErrorMsg, expression: Any = None):
        self.message = message.value
        self.expression = expression



import copy

import pymongo
from tqdm import tqdm

from tweepipe import settings
from tweepipe.db import db_client
from tweepipe.utils import snowflake


def _count_hashtag_over_tid_range(env_file, time_ranges, dbs, queries):
    """Get hashtag count over time ranges for multiple databases."""

    if env_file:
        settings.load_config(env_file=env_file)

    db_conn = db_client.DBClient()
    result_dict = {tr: {query["name"]: 0 for query in queries} for tr in time_ranges}

    for db in dbs:
        for time_range in tqdm(time_ranges):
            # recover approximative tid from date
            start_tid = snowflake.get_tweet_id_from_time(time_range[0])
            end_tid = snowflake.get_tweet_id_from_time(time_range[1])

            tid_range_query = {"tid": {"$gte": start_tid, "$lt": end_tid}}
            for query in queries:
                tmp_query = copy.deepcopy(query["query"])
                tmp_query.update(tid_range_query)

                tmp_collection = db_conn._get_collection(
                    db["collection"], db_name=db["db"]
                )

                # tmp_result = tmp_collection.count_documents(tmp_query)
                tmp_result = tmp_collection.find(
                    filter=tmp_query,
                    sort=[("tid", pymongo.ASCENDING)],
                    allow_disk_use=True,
                ).count()
                result_dict[time_range][query["name"]] += tmp_result

    return result_dict


def _aggregate_search_on_database():
    pass



import copy

import pymongo

from tweepipe import settings

DEFAULT_TWITTER_DB_SCHEMA = {
    "api": {"api": [{"index": "access_token", "unique": True}]},
    "default": {
        "tweets": [
            {"index": "tid", "unique": True},
            {"index": "uid", "unique": False},
        ],
        "users": [
            {"index": "uid", "unique": True},
            {"index": "screen_name", "unique": False},
        ],
        "hashtag_relations": [
            {"index": "tid", "unique": False},
            {"index": "user_id", "unique": False},
            {"index": [("hashtag", pymongo.TEXT)], "unique": False},
        ],
        "mention_relations": [
            {"index": "tid", "unique": False},
            {"index": "user_id", "unique": False},
            {"index": "mentionned_user_id", "unique": False},
        ],
        "retweet_relations": [
            {"index": "tid", "unique": False},
            {"index": "user_id", "unique": False},
            {"index": "retweet_id", "unique": False},
            {"index": "retweeted_user_id", "unique": False},
        ],
        "quote_relations": [
            {"index": "tid", "unique": False},
            {"index": "user_id", "unique": False},
            {"index": "quoted_user_id", "unique": False},
            {"index": "quoted_tweet_id", "unique": False},
        ],
        "reply_relations": [
            {"index": "tid", "unique": False},
            {"index": "user_id", "unique": False},
            {"index": "in_reply_to_user_id", "unique": False},
            {"index": "in_reply_to_tweet_id", "unique": False},
        ],
        "relation_extraction": [{"index": "tid", "unique": False}],
        "fetching_uids": [{"index": "tid", "unique": False}],
    },
}

INDEX_V3 = {
    "api": {"api": [{"index": "access_token", "unique": True}]},
    "default": {
        "tweets": [
            {"index": "tid", "unique": True},
            {"index": "json.created_at", "unique": False},
            {"index": "uid", "unique": False},
        ],
        "users": [
            {"index": "uid", "unique": True},
            {"index": "screen_name", "unique": False},
        ],
        "hashtag_relations": [
            {"index": "tid", "unique": False},
            {"index": "created_at", "unique": False},
            {"index": "user_id", "unique": False},
            {"index": [("hashtag", pymongo.TEXT)], "unique": False},
        ],
        "mention_relations": [
            {"index": "tid", "unique": False},
            {"index": "created_at", "unique": False},
            {"index": "user_id", "unique": False},
            {"index": "mentionned_user_id", "unique": False},
        ],
        "retweet_relations": [
            {"index": "tid", "unique": False},
            {"index": "created_at", "unique": False},
            {"index": "user_id", "unique": False},
            {"index": "retweet_id", "unique": False},
            {"index": "retweeted_user_id", "unique": False},
        ],
        "quote_relations": [
            {"index": "tid", "unique": False},
            {"index": "created_at", "unique": False},
            {"index": "user_id", "unique": False},
            {"index": "quoted_user_id", "unique": False},
            {"index": "quoted_tweet_id", "unique": False},
        ],
        "reply_relations": [
            {"index": "tid", "unique": False},
            {"index": "created_at", "unique": False},
            {"index": "user_id", "unique": False},
            {"index": "in_reply_to_user_id", "unique": False},
            {"index": "in_reply_to_tweet_id", "unique": False},
        ],
        "relation_extraction": [{"index": "tid", "unique": False}],
        "fetching_uids": [
            {"index": "uid", "unique": True},
            {"index": "status", "unique": False},
        ],
        "hydrating_tids": [
            {"index": "tid", "unique": True},
            {"index": "key", "unique": False},
            {"index": "created_at", "unique": False},
        ],
        "likes": [
            {"index": "tid", "unique": True},
        ],
    },
}

INDEX_V3 = {
    "api": {"api": [{"index": "access_token", "unique": True}]},
    "default": {
        "tweets": [
            {"index": "tid", "unique": True},
            {"index": "json.created_at", "unique": False},
            {"index": "uid", "unique": False},
        ],
        "users": [
            {"index": "uid", "unique": True},
            {"index": "screen_name", "unique": False},
        ],
        "hashtags": [
            {"index": "tid", "unique": False},
            {"index": "created_at", "unique": False},
            {"index": "user_id", "unique": False},
            {"index": [("hashtag", pymongo.TEXT)], "unique": False},
        ],
        "mentions": [
            {"index": "tid", "unique": False},
            {"index": "created_at", "unique": False},
            {"index": "user_id", "unique": False},
            {"index": "mentionned_user_id", "unique": False},
        ],
        "retweets": [
            {"index": "tid", "unique": False},
            {"index": "created_at", "unique": False},
            {"index": "user_id", "unique": False},
            {"index": "retweet_id", "unique": False},
            {"index": "retweeted_user_id", "unique": False},
        ],
        "quotes": [
            {"index": "tid", "unique": False},
            {"index": "created_at", "unique": False},
            {"index": "user_id", "unique": False},
            {"index": "quoted_user_id", "unique": False},
            {"index": "quoted_tweet_id", "unique": False},
        ],
        "replies": [
            {"index": "tid", "unique": False},
            {"index": "created_at", "unique": False},
            {"index": "user_id", "unique": False},
            {"index": "in_reply_to_user_id", "unique": False},
            {"index": "in_reply_to_tweet_id", "unique": False},
        ],
        "relation_extraction": [{"index": "tid", "unique": False}],
        "fetching_uids": [
            {"index": "uid", "unique": True},
            {"index": "status", "unique": False},
        ],
        "hydrating_tids": [
            {"index": "tid", "unique": True},
            {"index": "key", "unique": False},
            {"index": "created_at", "unique": False},
        ],
        "likes": [
            {"index": "tid", "unique": True},
        ],
    },
}

VOTECOMPASS_DATA = {
    "default": {
        "linked_tweets": [
            {"index": "tid", "unique": True},
            {"index": "json.created_at", "unique": False},
            {"index": "uid", "unique": False},
            {"index": "status", "unique": False},
        ],
        "votecompass_result": [
            {"index": "tid", "unique": False},
            {"index": "uid", "unique": False},
            {"index": "url", "unique": False},
            {"index": "filepath", "unique": False},
        ],
    }
}


USER_LOOKUP_V1 = {
    "default": {
        "users": [
            {"index": "uid", "unique": True},
            {"index": "screen_name", "unique": False},
        ],
        "missing_users": [{"index": "uid", "unique": True}],
    }
}

BOTSPOT_V1 = {
    "default": {
        "users": [
            {"index": "uid", "unique": True},
            {"index": "username", "unique": False},
        ],
    }
}

TWITTER_RELATION_TYPES = [
    "retweet_relations",
    "hashtag_relations",
    "mention_relations",
    "quote_relations",
    "reply_relations",
]

DEFAULT_ACADEMIC_V2_SCHEMA = {
    "default": {
        "places": [{"index": "id", "unique": True}],
        "polls": [{"index": "id", "unique": True}],
        "media": [{"index": "media_key", "unique": True}],
        "tweets": [{"index": "id", "unique": True}],
        "users": [{"index": "id", "unique": True}],
        "fetching_uids": [
            {"index": "uid", "unique": True},
            {"index": "status", "unique": False},
        ],
    }
}

VOTECOMPASS_THUMBNAIL = {
    "default": {
        "places": [{"index": "id", "unique": True}],
        "polls": [{"index": "id", "unique": True}],
        "media": [{"index": "media_key", "unique": True}],
        "thumbnails": [{"index": "id", "unique": False}],
        "tweets": [{"index": "id", "unique": True}],
        "users": [{"index": "id", "unique": True}],
    }
}

LIKES_DB = {"likes": [{"index": "tid", "unique": False}]}


def _get_issue_name_from_dt(start_time, end_time, base_issue):
    """Format name based on str tmestamp."""

    start_time = "".join(i for i in start_time if i.isdigit())
    end_time = "".join(i for i in end_time if i.isdigit())

    return "{}_{}_to_{}".format(base_issue, start_time, end_time)


def _get_collection_name_from_dt(start_time: str, end_time: str):
    """Get a mirror response collection name based on two datetimes.

    Implemented for debugging purposes, will likely end up as deadcode.

    """

    collection_name = settings.RESPONSE_MIRROR_COLLECTION_NAME
    collection_name += "_{}_to_{}".format(
        "".join(i for i in start_time if i.isdigit()),
        "".join(i for i in end_time if i.isdigit()),
    )

    return collection_name


def _get_db_schema(
    issue: str = None,
    include_users: bool = True,
    include_relations: bool = True,
    schema_base: dict = DEFAULT_TWITTER_DB_SCHEMA,
):

    _db_schema = copy.deepcopy(schema_base)

    if issue:
        _db_schema[issue] = _db_schema["default"]

        if not include_users:
            del _db_schema[issue]["users"]

    return _db_schema



import time

import tweepy
from loguru import logger

from tweepipe import settings
from tweepipe.db import db_schema, db_client


def _get_processed_users(lookup_batch):
    return {user._json["id_str"]: user._json for user in lookup_batch}


def _lookup_users(batch, twitter_api):
    try:
        lookedup_users = twitter_api.lookup_users(user_id=batch)
    except Exception as e:
        print(f"Got following error: {e}.")
        time.sleep(3)
        return {}

    return _get_processed_users(lookedup_users)


def _lookup_users_from_uids(
    uids: list, twitter_api: tweepy.API, issue: str, env_file: str = None
):
    if env_file:
        settings.load_config(env_file=env_file)

    # setup target database with collections
    db_conn = db_client.DBClient(issue=issue, schema=db_schema.USER_LOOKUP_V1)

    missing_user_docs = []
    missing_user_count = 0
    retrieved_user_count = 0

    lookup_batch_size = 100
    for i in range(0, len(uids), lookup_batch_size):
        result = _lookup_users(uids[i : i + lookup_batch_size], twitter_api)
        docs = [{"json": result[u], "uid": u} for u in result]
        db_conn.add_users(docs)
        retrieved_user_count += len(docs)

        tmp_missing_user_docs = [{"uid": uid} for uid in result if uid not in result]
        missing_user_docs.extend(tmp_missing_user_docs)
        missing_user_count += len(tmp_missing_user_docs)

    db_conn.add_missing_users(missing_user_docs)
    logger.info(
        f"Retrieved {retrieved_user_count} users, found {missing_user_count} missing users."
    )
    logger.info(
        f"Total, got {float(retrieved_user_count/(retrieved_user_count+missing_user_count)):.2f}% of users for this batch."
    )



import datetime
from concurrent.futures import thread

import semver
import tweepy
from loguru import logger

# make sure tweepy 4+ is available in env
assert semver.VersionInfo.parse(tweepy.__version__).major >= 4

from tweepipe import settings
from tweepipe.db import db_client, db_schema
from tweepipe.utils import streaming_client


class AcademicClient:
    """Tweepy Client dedicated to handling operations with the new
    Academic API. Compatible with tweepy 4.0.0 and above.
    """

    def __init__(
        self,
        issue: str,
        include_users: bool = True,
        include_relations: bool = True,
        schema: dict = db_schema.DEFAULT_ACADEMIC_V2_SCHEMA,
        env_file: str = None,
        stream: bool = False,
    ):
        self._env = env_file
        self.issue = issue

        self.db_conn = db_client.DBClient(
            issue=issue,
            schema=schema,
            include_relations=include_relations,
            include_users=include_users,
        )
        if not stream:
            self.tweepy_client = tweepy.Client(
                bearer_token=settings.ACADEMIC_API_BEARER_TOKEN, wait_on_rate_limit=True
            )
            self.streaming_client = None
        else:
            self.streaming_client = streaming_client.StreamingClient(
                bearer_token=settings.ACADEMIC_API_BEARER_TOKEN,
                db_conn=self.db_conn,
                wait_on_rate_limit=True,
            )
            self.tweepy_client = None

    def create_rule(self, keywords):
        if not self.streaming_client:
            raise RuntimeError("Streaming client not initialized")

        # add a rule for filtering the sample stream
        rule = tweepy.StreamRule(" ".join(keywords))
        print("Creating rule", rule)
        self.streaming_client.add_rules(rule, dry_run=True)

    def get_likes(self, uids: list, since_ts: datetime.datetime):
        # log all uids as being retrieved
        processed_uid_docs = [{"uid": uid, "status": 0} for uid in uids]
        self.db_conn.add_fetching_uids(processed_uid_docs)
        if len(self.db_conn.fetching_uids_batch) > 0:
            self.db_conn._push_fetching_uids_to_db()

        for uid in uids:
            self.db_conn.db_conn.update_one({"uid": uid}, {"$set": {"status": 1}})

            # keep going if error occurs
            try:
                likes = self.tweepy_client.get_likes(
                    uid,
                )
            except:
                pass

        return

    def stream(self):
        if not self.streaming_client:
            raise RuntimeError("Streaming client not initialized")

        # start streaming client
        self.streaming_client.sample(
            backfill_minutes=5,
            expansions=self._get_expansions(),
            user_fields=self._get_user_fields(),
            tweet_fields=self._get_user_fields(),
            threaded=True,
        )

        # not storing these for now
        # place_fields=self._get_place_fields(),
        # poll_fields=self._get_poll_fields(),

    def search_user_history(
        self,
        uid: list,
        since_ts: datetime.datetime,
        end_ts: datetime.datetime = datetime.datetime.now(),
        keywords: list = None,
    ):
        search_query = self._get_query(keywords=keywords, uid=uid)
        fetching_uids_collection = self.db_conn._get_collection(
            "fetching_uids", db_name=self.issue
        )
        fetching_uids_collection.update_one({"uid": uid}, {"$set": {"status": 1}})

        kwargs = dict(
            query=search_query,
            start_time=since_ts,
            end_time=end_ts,
            expansions=self._get_expansions(),
            max_results=100,
            media_fields=self._get_media_fields(),
            place_fields=self._get_place_fields(),
            tweet_fields=self._get_tweet_fields(),
            poll_fields=self._get_poll_fields(),
            user_fields=self._get_user_fields(),
        )
        found_tweets = self._paginate_tweets(
            self.tweepy_client.search_all_tweets, kwargs
        )
        fetching_uids_collection.update_one(
            {"uid": uid}, {"$set": {"status": 2 if found_tweets else -1}}
        )

    def search_all(
        self,
        keywords: list,
        start_time: datetime,
        end_time: datetime,
        tweets_only: bool = False,
        max_results: int = 100,
    ):
        """Run an all time search with the Twitter API."""

        search_query = self._get_query(
            keywords=keywords,
            exclude_retweet=tweets_only,
            exclude_reply=tweets_only,
            exclude_quote=tweets_only,
        )

        tweepy_client = tweepy.Client(
            bearer_token=settings.ACADEMIC_API_BEARER_TOKEN, wait_on_rate_limit=True
        )
        kwargs = dict(
            query=search_query,
            start_time=start_time,
            end_time=end_time,
            expansions=self._get_expansions(),
            max_results=max_results,
            media_fields=self._get_media_fields(),
            place_fields=self._get_place_fields(),
            poll_fields=self._get_poll_fields(),
            tweet_fields=self._get_tweet_fields(),
            user_fields=self._get_user_fields(),
        )
        _ = self._paginate_tweets(fn=tweepy_client.search_all_tweets, kwargs=kwargs)

    def _paginate_tweets(self, fn, kwargs):
        """Iterate through all pages returned by arg function."""

        response_count, retrieved_tweet_count = 0, 0
        for response in tweepy.Paginator(fn, **kwargs):
            retrieved_tweet_count += self.save_response(response)
            response_count += 1
            if response_count % 100 == 0:
                logger.info(
                    f"Retrieved {retrieved_tweet_count} tweets, {response_count} responses."
                )

        # make sure all fresh data is pushed to the database
        self.db_conn.flush_content()

        logger.info(
            f"Retrieved a total of {retrieved_tweet_count} tweets, {response_count} responses."
        )

        return response_count > 0

    def push_all(self):
        if len(self.db_conn.media_batch) > 0:
            self.db_conn._push_media_to_db()
        if len(self.db_conn.place_batch) > 0:
            self.db_conn._push_place_to_db()
        if len(self.db_conn.poll_batch) > 0:
            self.db_conn._push_poll_to_db()
        if len(self.db_conn.tweet_batch) > 0:
            self.db_conn._push_tweets_to_db()
        if len(self.db_conn.user_batch) > 0:
            self.db_conn._push_users_to_db()

    def save_response(self, response: tweepy.Response) -> int:
        """Save each response data item returned by server."""

        retrieved_tweet_count = 0
        try:
            for tweet in response.data:
                retrieved_tweet_count += 1
                try:
                    self.db_conn.add_tweet(tweet.data)
                except Exception as e:
                    logger.error(f"Error adding tweet to db. {e}")
        except Exception as e:
            logger.error(f"Error saving response. {e}")
            return

        if response.includes.get("media"):
            for media in response.includes.get("media"):
                try:
                    self.db_conn.add_media(media.data)
                except Exception as e:
                    continue

        if response.includes.get("users"):
            for user in response.includes.get("users"):
                try:
                    self.db_conn.add_user(user.data)
                except Exception as e:
                    continue

        if response.includes.get("places"):
            for place in response.includes.get("places"):
                try:
                    self.db_conn.add_place(place.data)
                except Exception as e:
                    continue

        if response.includes.get("polls"):
            for poll in response.includes.get("polls"):
                try:
                    self.db_conn.add_poll(poll.data)
                except Exception as e:
                    continue

        return retrieved_tweet_count

    def _get_query(
        self,
        keywords: list,
        uid: str = None,
        exclude_retweet: bool = False,
        exclude_quote: bool = False,
        exclude_reply: bool = False,
        query: str = "",
    ):
        """Build Academic API search query."""

        # include keywords, assuming empty query
        if keywords:
            query += " OR ".join(keywords)
            query = "( " + query + " )"

        if uid:
            query += f" from:{uid}"

        # check if should exclude queries or not
        if exclude_retweet:
            query += " -is:retweet "

        # if exclude_quote:
        #    query += " -is:quote "

        # if exclude_reply:
        #    query += " -is:reply "

        # include image requirement by default
        # query += " has:media "

        # query += " has:images "

        # has:images or has:media

        return query.strip()

    def _get_poll_fields(self):
        poll_fields = [
            "duration_minutes",
            "end_datetime",
            "id",
            "options",
            "voting_status",
        ]
        return poll_fields

    def _get_place_fields(self):
        place_fields = [
            "contained_within",
            "country",
            "country_code",
            "full_name",
            "geo",
            "id",
            "name",
            "place_type",
        ]
        return place_fields

    def _get_media_fields(self):
        media_fields = [
            "duration_ms",
            "media_key",
            "preview_image_url",
            "type",
            "url",
            "public_metrics",
            "alt_text",
        ]
        return media_fields

    def _get_tweet_fields(self):
        tweet_fields = [
            "id",
            "text",
            "attachments",
            "author_id",
            "context_annotations",
            "conversation_id",
            "created_at",
            "entities",
            "geo",
            "in_reply_to_user_id",
            "lang",
            "public_metrics",
            "possibly_sensitive",
            "referenced_tweets",
        ]

        return ",".join(tweet_fields)

    def _get_user_fields(self):
        user_fields = [
            "id",
            "name",
            "username",
            "created_at",
            "description",
            "entities",
            "location",
            "pinned_tweet_id",
            "profile_image_url",
            "protected",
            "public_metrics",
            "url",
            "verified",
        ]

        return ",".join(user_fields)

    def _get_expansions(self):
        expansions = [
            "author_id",
            "referenced_tweets.id",
            "in_reply_to_user_id",
            "attachments.media_keys",
            "attachments.poll_ids",
            "geo.place_id",
            "entities.mentions.username",
            "referenced_tweets.id.author_id",
        ]

        return ",".join(expansions)



from tweepipe import settings
from tweepipe.legacy import client
from tweepipe.legacy.botspot import botspot

settings.load_config(env_file=".env")


def main():
    # define our list of uids
    uids = ["800111181058838528", "1284651818782535680", "859219800"]

    issue = "my_issue"
    my_client = client.LegacyClient()
    profiles = my_client.get_user_profiles(uids, db=issue)

    # retrieve profile scores
    my_botspot = botspot.BotSpot()
    print("Retrieved scores for the following users:")
    for item in my_botspot.get_users_scores(profiles).items():
        print(f"User {item[0]}: {item[1]}")


if __name__ == "__main__":
    main()



import datetime
import json
import random
import requests
import semver
from tweepipe.legacy.utils import extract


random.seed(123)

import searchtweets
import yaml
import tweepy
from tqdm import tqdm
import pymongo
from loguru import logger
from omegaconf import OmegaConf

# can only import streamlistener if tweepy is previous to 4.0
if semver.VersionInfo.parse(tweepy.__version__).major < 4:
    from tweepipe import streaming

from tweepipe import hydrating, settings, searching, lookup
from tweepipe.db import db_client, db_schema
from tweepipe.utils import (
    credentials,
    loader,
    parallel,
    streaming_client,
)
from tweepipe.base import history
from tweepipe.utils.migration import convert, versions
from tweepipe.utils.snowflake import SnowFlake


class LegacyClient:
    """
    Legacy client to work with Tweepy < 4.0.0. Compatible with the tweepipe MongoDB client.

    :param issue: Name of the ongoing issue, also the working database. Defaults to None.
    :type issue: str, optional
    :param include_users: Whether to extract user profiles upon processing the retrieved tweets.
        If so, the users will be stored in a separate collection. Defaults to True.
    :type include_users: bool, optional
    :param include_relations: Whether to extract relations (hashtags, quotes, mentions, retweets)
        upon processing. Like users, if extracted they will be stored under separate collections
        in the mongo database. Defaults to True.
    :type include_relations: bool, optional
    :param skip_db: Set to True to skip usage of a MongoDB, and not have persistent storage on
        collection tasks. Defaults to False.
    :type skip_db: bool, optional
    :param schema: Specify an index schema for the retrieved Twitter data. For example, an index
        on tweet ids can be useful to quickly get to a tweet. Defaults to db_schema.INDEX_V3.
    :type schema: dict, optional
    """

    def __init__(
        self,
        issue: str = None,
        include_users: bool = True,
        include_relations: bool = True,
        skip_db: bool = False,
        schema: dict = None,
        stream: bool = False,
        api_credentials: dict = None,
    ):

        self._issue = issue
        self._include_users = include_users
        self._include_relations = include_relations
        self._schema = schema if schema else db_schema.INDEX_V3

        try:
            if not skip_db:
                self._db_conn = db_client.DBClient(
                    issue=self._issue,
                    include_users=self._include_users,
                    include_relations=self._include_relations,
                    schema=self._schema,
                )
            else:
                self._db_conn = None
        except Exception as e:
            logger.error(f"Encountered error establishing db connection {e}")
            self._db_conn = None
            pass

        self._env_file = None

        if stream and not skip_db:
            # get credentials from environment
            # TODO: link to mongodb to pull credentials from there

            if not api_credentials:
                api_credentials = dict(
                    consumer_key=settings.CONSUMER_KEY,
                    consumer_secret=settings.CONSUMER_SECRET,
                    access_token=settings.ACCESS_TOKEN,
                    access_token_secret=settings.ACCESS_TOKEN_SECRET,
                )

            self.streaming_client = streaming_client.StreamV1(
                db_conn=self._db_conn, output_file="stream.json", **api_credentials
            )
        else:
            self.streaming_client = None

    def stream(self, keywords: list = None, languages: list = None):
        if not self.streaming_client:
            raise RuntimeError("Streaming client not initialized")

        # run a sampling stream, directed towards db
        if keywords:
            self.streaming_client.filter(track=keywords)
        elif languages:
            self.streaming_client.sample(languages=languages)
        else:
            self.streaming_client.sample()

    def create_rule(self, keywords):
        if not self.streaming_client:
            raise RuntimeError("Streaming client not initialized")

        # add a rule for filtering the sample stream
        rule = tweepy.StreamRule(" ".join(keywords))
        self.streaming_client.add_rules(rule)

    def stream_twitter_feed_seq(
        self,
        keywords: list = [],
        keywords_file: str = None,
        issue: str = None,
        twitter_credentials: dict = None,
        output_file: str = None,
        skip_db: bool = False,
    ):
        if len(keywords) == 0 and keywords_file:
            keywords = loader._load_from_file(keywords_file)

        tmp_issue = self._get_issue(issue=issue)

        if not twitter_credentials:
            twitter_credentials = credentials._get_twitter_credentials(
                db_conn=self.db_conn, api_count=1, purpose="stream"
            )[0]

        logger.info(
            f"Retrieved {twitter_credentials['consumer_key']} twitter authentication for sequential streaming."
        )

        streaming.stream_keywords(
            keywords=keywords,
            twitter_credentials=twitter_credentials,
            issue=tmp_issue,
            output_file=output_file,
            include_users=self.include_users,
            include_relations=self.include_relations,
            skip_db=skip_db,
        )

    def get_user_profiles(self, uids: list, db: str, collection: str = "users"):
        """Retrieve user profiles."""

        user_profiles = []
        target_collection = self.db_conn._get_collection(collection, db_name=db)
        projection = {"_id": False}
        for uid in uids:
            query = {"uid": uid}
            tmp_user_profile = target_collection.find_one(
                filter=query, projection=projection
            )
            if tmp_user_profile:
                user_profiles.append(tmp_user_profile.get("json"))

        return user_profiles

    def stream_twitter_feed(
        self,
        keywords: list = [],
        keywords_file: str = None,
        issue: str = None,
        api_count: int = 1,
        twitter_credentials: list = None,
    ):
        """Stream twitter feed using Twitter APIs. The default execution uses a single process.

        To perform parallel streaming, specify an api_count greater than 1. To employ all
            available apis, set api_count to -1.
        """

        if len(keywords) == 0 and keywords_file:
            keywords = loader._load_from_file(keywords_file)

        # retrieve issue + twitter credentials
        tmp_issue = self._get_issue(issue=issue)
        if not twitter_credentials:
            twitter_credentials = credentials._get_twitter_credentials(
                db_conn=self.db_conn, api_count=api_count, purpose="stream"
            )
        twitter_auths = credentials._get_twitter_auths(twitter_credentials)
        logger.info(f"Retrieved {len(twitter_auths)} twitter authentication insts.")

        # run parallel streamers by batching keywords
        process_count = len(twitter_auths)
        kwargs_list = streaming._get_parallel_streaming_kwargs(
            keywords=keywords,
            twitter_auths=twitter_auths,
            process_count=process_count,
            issue=tmp_issue,
            include_users=self.include_users,
            include_relations=self.include_relations,
        )
        results = parallel.run_parallel(
            fn=streaming._run_streamer,
            kwargs_list=kwargs_list,
            max_workers=process_count,
        )

    def get_likes(self, uids: list, since_ts: datetime.datetime, api_count: int = 1):
        # log all uids as being retrieved
        processed_uid_docs = [{"uid": uid, "status": 0} for uid in uids]
        self.db_conn.add_fetching_uids(processed_uid_docs)
        if len(self.db_conn.fetching_uids_batch) > 0:
            self.db_conn._push_fetching_uids_to_db()

        twitter_apis, twitter_credentials = credentials._get_twitter_apis(
            db_conn=self.db_conn,
            api_count=1,
            purpose="hydrate",
            free_api=True,
        )
        twitter_api = twitter_apis[0]

        since_tid = SnowFlake.get_tweet_id_from_time(since_ts)

        fetching_uids = self.db_conn._get_collection("fetching_uids")

        for uid in tqdm(uids):
            fetching_uids.update_one({"uid": uid}, {"$set": {"status": 1}})
            # keep going if error occurs
            try:
                for page in tweepy.Cursor(
                    twitter_api.favorites,
                    id=uid,
                    count=100,
                    since_id=since_tid,
                    tweet_mode="extended",
                ).pages():
                    for status in page:
                        extract.extract_likes(uid, status._json, self.db_conn)
                fetching_uids.update_one({"uid": uid}, {"$set": {"status": 2}})
            except Exception as e:
                print(e)
                fetching_uids.update_one({"uid": uid}, {"$set": {"status": -1}})
                pass

        self.db_conn.flush_content()

        return

    def get_user_friends(
        self,
    ):
        """Retrieve friends for a single user."""

    def get_users_friends(
        self,
    ):
        """Retrieve friends for multiple users."""

    def fetch_followers(
        self,
        uids: list,
        free_api: bool = False,
        api_count: int = 1,
        cap_count: int = 50000,
    ):
        pass

    def load_tweet_files_to_hydrate(
        self, files: list, batch_size: int = 4096, worker_count: int = 6
    ):
        """Load tweet ids from files into collection for hydration.

        Use a status to get tids to hydrate:
            0: waiting
            1: hydrating
            2: done
            -1: missing
        """

        # make sure hydrating index exists
        self.db_conn.create_hydrating_index()

        kwargs_list = loader._get_loading_tids_kwargs(
            files=files,
            batch_size=batch_size,
            env_file=self.env_file,
            issue=self._get_issue(),
            worker_count=worker_count,
        )

        # loader.load_tids_from_file_to_db(**kwargs_list.pop())
        results = parallel.run_parallel(
            fn=loader.load_tids_from_file_to_db,
            kwargs_list=kwargs_list,
            max_workers=worker_count,
        )

    def hydrate_tweets_from_db(
        self,
        start_date: datetime.datetime = None,
        end_date: datetime.datetime = None,
        api_count: int = 1,
        issue: str = None,
        target_db: str = None,
        batch_size: int = 1024,
        include_relations: bool = True,
        include_users: bool = True,
        collection: str = "hydrating_tids",
        free_api: bool = False,
    ):
        """Same as the hydrate tweets method, however take all tids from
            database instead of taking them from a file or input list.

        Note that the current hydrating tids for the usc elections database
            use a field tid_int which stores all tweet ids as integers additionally
            to the tid field which stores the tweet ids as strings. All additional
            id fields are now moved to integers for simpler searches.
        """

        include_users = self._get_include_users(include_users)
        include_relations = self._get_include_relations(include_relations)

        tmp_issue = self._get_issue(issue=issue)
        target_db = tmp_issue if not target_db else target_db
        logger.info(
            f"Hydrating data from {tmp_issue}. Include users: {include_users}. Include relations {include_relations}."
        )

        twitter_apis, twitter_credentials = credentials._get_twitter_apis(
            db_conn=self.db_conn,
            api_count=api_count,
            purpose="hydrate",
            free_api=free_api,
        )
        logger.info(f"Retrieved {len(twitter_apis)} api accesses.")
        if len(twitter_apis) == 1:
            twitter_api = twitter_apis.pop()
            logger.info(
                f"Starting hydrating using API access with key {twitter_credentials[0].get('consumer_key')}."
            )
            results = hydrating._run_hydration_from_db(
                twitter_api=twitter_api,
                include_relations=include_relations,
                include_users=include_users,
                target_db=target_db,
                issue=tmp_issue,
                collection=collection,
                env_file=self.env_file,
                batch_size=batch_size,
                start_date=start_date,
                end_date=end_date,
            )
        else:
            kwargs_list = hydrating._get_parallel_hydrating_from_db_kwargs(
                twitter_apis=twitter_apis,
                issue=tmp_issue,
                target_db=target_db,
                include_users=include_users,
                include_relations=include_relations,
                env_file=self.env_file,
                collection=collection,
                start_date=start_date,
                end_date=end_date,
            )
            results = parallel.run_parallel(
                fn=hydrating._run_hydration_from_db,
                kwargs_list=kwargs_list,
                max_workers=len(twitter_apis),
            )

        return results

    def hydrate_tweets(
        self,
        tweet_ids: list = [],
        tweet_ids_file: str = None,
        issue: str = None,
        api_count: int = 1,
    ):
        """Recover tweet contents from tweet ids."""

        if tweet_ids_file:
            tweet_ids = loader._load_from_file(tweet_ids_file)

        tmp_issue = self._get_issue(issue=issue)

        # retrieve twitter apis
        twitter_apis, _ = credentials._get_twitter_apis(
            db_conn=self.db_conn, api_count=api_count, purpose="hydrate"
        )

        # run sequential hydration process
        if len(twitter_apis) == 1:
            twitter_api = twitter_apis.pop()
            results = hydrating._run_hydration(
                twitter_api=twitter_api,
                tweet_ids=tweet_ids,
                issue=tmp_issue,
                include_users=self.include_users,
                include_relations=self.include_relations,
            )
        # run parallel streamers by batching keywords
        else:
            process_count = len(twitter_apis)
            kwargs_list = hydrating._get_parallel_hydrating_kwargs(
                tweet_ids=tweet_ids,
                twitter_apis=twitter_apis,
                issue=tmp_issue,
                env_file=self.env_file,
                include_relations=self.include_relations,
                include_users=self.include_users,
            )
            # in results is returned lists of hydrated tweets
            results = parallel.run_parallel(
                fn=hydrating._run_hydration,
                kwargs_list=kwargs_list,
                max_workers=process_count,
            )

        return results

    def fetch_user_history(self, uid):
        """Recover a user's tweets timeline."""

    def search_users_histories(
        self,
        uids: list,
        since_ts: datetime.datetime,
        end_ts: datetime.datetime = datetime.datetime.now(),
        issue: str = None,
        keywords: list = None,
        include_users: bool = True,
        include_relations: bool = True,
    ):
        issue = self._get_issue(issue=issue)
        include_users = self._get_include_users(include_users=include_users)
        include_relations = self._get_include_relations(
            include_relations=include_relations
        )

        # log all uids to be scraped in this process
        processed_uid_docs = [{"uid": uid, "status": 0} for uid in uids]
        self.db_conn.add_fetching_uids(processed_uid_docs)
        if len(self.db_conn.fetching_uids_batch) > 0:
            self.db_conn._push_fetching_uids_to_db()

        valid_uid_updates, invalid_uid_updates = [], []

        # get collection for uids being fetched
        fetching_uids_collection = self.db_conn._get_collection(
            "fetching_uids", db_name=issue
        )
        to_fetch_uids = list(fetching_uids_collection.find(filter={"status": 0}))
        logger.info(f"Searching for {len(to_fetch_uids)} uids using Academic API.")
        academic_credentials = self._get_academic_credentials()
        for uid in tqdm(to_fetch_uids):
            raw_output_filename, output_filename = self._get_output_filepaths(
                start_time=since_ts, end_time=end_ts, uid=uid
            )
            try:
                logger.info(f"Searching for {uid['uid']}")
                found_tweets = self._run_full_archive_continuous_search(
                    raw_output_filename=raw_output_filename,
                    output_filename=output_filename,
                    credentials=academic_credentials,
                    uid=uid["uid"],
                    keywords=keywords,
                    start_time=since_ts,
                    end_time=end_ts,
                )
            except requests.exceptions.HTTPError as e:
                print(e)
                logger.error(e)
                found_tweets = False

            if found_tweets:
                fetching_uids_collection.update_one(
                    {"uid": uid}, {"$set": {"status": 2}}
                )
            else:
                fetching_uids_collection.update_one(
                    {"uid": uid}, {"$set": {"status": -1}}
                )

    def fetch_users_from_db_histories(
        self,
        since_ts: datetime.datetime = None,
        api_count: int = 1,
        free_api: bool = False,
        api_purpose: str = "lookup",
    ):
        logger.info(
            f"Starting lookup process on {self.issue}. Include relations: {self.include_relations}. Include users: {self.include_users}."
        )
        twitter_apis, twitter_credentials = credentials._get_twitter_apis(
            db_conn=self.db_conn,
            api_count=api_count,
            purpose=api_purpose,
            free_api=free_api,
        )

        kwargs_list = searching._get_fetch_from_db_histories_kwargs(
            twitter_apis=twitter_apis,
            since_ts=since_ts,
            issue=tmp_issue,
            include_users=include_users,
            include_relations=include_relations,
        )

        results = parallel.run_parallel(
            fn=searching._fetch_history,
            kwargs_list=kwargs_list,
            max_workers=len(twitter_apis),
        )

        credentials.free_api(
            db_conn=self.db_conn,
            twitter_credentials=twitter_credentials,
            purpose=api_purpose,
        )

    def fetch_users_histories(
        self,
        uids: list,
        since_ts: datetime.datetime,
        api_count: int = -1,
        issue: str = None,
        free_api: bool = False,
        include_users: bool = True,
        include_relations: bool = True,
        twitter_credentials: list = None,
        api_purpose: str = "lookup",
    ):
        """Recover multiple users tweets timelines."""

        include_users = self._get_include_users(include_users)
        include_relations = self._get_include_relations(include_relations)
        tmp_issue = self._get_issue(issue=issue)
        logger.info(
            f"Starting lookup process on {tmp_issue}. Include relations: {include_relations}. Include users: {include_users}."
        )
        twitter_apis, twitter_credentials = credentials._get_twitter_apis(
            db_conn=self.db_conn,
            api_count=api_count,
            purpose=api_purpose,
            twitter_credentials=twitter_credentials,
            free_api=free_api,
        )

        logger.info(
            f"Looking up {len(uids)} users with {[cred['consumer_key'] for cred in twitter_credentials]} apis."
        )

        kwargs_list = searching._get_fetch_histories_kwargs(
            twitter_apis=twitter_apis,
            twitter_credentials=twitter_credentials,
            uids=uids,
            since_ts=since_ts,
            issue=tmp_issue,
            include_users=include_users,
            include_relations=include_relations,
        )

        results = parallel.run_parallel(
            fn=searching._fetch_history,
            kwargs_list=kwargs_list,
            max_workers=len(twitter_apis),
        )

        credentials.free_api(
            db_conn=self.db_conn,
            twitter_credentials=twitter_credentials,
            purpose=api_purpose,
        )

    def lookup_users_sequential(
        self,
        uids: list = [],
        issue: str = None,
        api_credentials: dict = None,
        env_file: str = None,
    ):
        twitter_api = credentials._get_twitter_api(credentials=api_credentials)
        print("Collecting data from Twitter API.")
        lookup._lookup_users_from_uids(
            uids=uids, twitter_api=twitter_api, issue=issue, env_file=env_file
        )

    def lookup_users(self, uids: list = [], issue: str = None, api_count: int = -1):
        """Lookup user profiles and insert them in database."""

        tmp_issue = self._get_issue(issue=issue)
        twitter_apis, _ = credentials._get_twitter_apis(
            db_conn=self.db_conn, api_count=api_count, purpose="lookup"
        )
        kwargs_list = parallel.get_lookup_users_kwargs_list(
            uids=uids,
            apis=twitter_apis,
            env_file=self.env_file,
            issue=tmp_issue,
        )
        logger.info(
            f"Preparing data retrieval for {len(kwargs_list)} user batches and {len(twitter_apis)} twitter apis."
        )
        results = parallel.run_parallel(
            fn=lookup._lookup_users_from_uids,
            kwargs_list=kwargs_list,
            max_workers=len(twitter_apis),
        )

    def check_user_activity(
        self,
    ):
        """Verify whether or not a user is active."""

    def check_users_activity(
        self,
    ):
        """Verify whether or not multiple users are active."""

    def get_botspot_score(
        self,
    ):
        """Score bot-likelihood of a user using botspot."""

    def lookup_tweets(self):
        """Basically the same thing as hydrate tweet but only for a single tweet id."""

    def recover_tweets(
        self,
        keywords: list = [],
        start_time: datetime.datetime = None,
        end_time: datetime.datetime = None,
        use_academic_api: bool = True,
        twitter_credentials: list = None,
        sampling_rate: int = 1,
        full: bool = False,
        bbox: str = None,
        country: str = None,
    ):
        """Search tweets from a given period of time.

        Args:
            keywords:       Keywords to use for data recovery.
            start_time: Datetime timestamp designating begining of missing data period.
            end_time:   Datetime timestamp designating end of missing data period.
            use_academic_api:       Request access to an academic api to perform the tweet
                    retrieval. This is required if getting tweets more than a week after they
                    were posted (limit of the standard api).
            expected_proportion: Depending on the purpose of the tweets recovery, one may
                want to reproduce the filtering process used by Twitter upon streaming tweets
                with the free API (yielding 1% of the total stream).

        If the client wishes to use an academic API (early access v2 at this point), the full
            archive search api will return a complete set of tweets corresponding to the query
            for the given time period. The result of this query will then be stored and saved
            for later processing. If the client is striving to complete a dataset, we assume that
            1% of the retrieved tweets are actually missing. We thus, sample the database of
            retrieved tweets to obtain a representative set of tweets.

        """

        if not use_academic_api and not self._is_valid_timestamp(start_time):
            raise ValueError("The timestamp goes back more than 7 days!")

        # use the searchtweets handler to query twitter's full archive
        if use_academic_api:
            self._full_archive_tweets_search(
                keywords=keywords,
                start_time=start_time,
                end_time=end_time,
                onfly_sampling=(not full),
                sampling_rate=sampling_rate,
                bbox=bbox,
                country=country,
            )
        # otherwise implement recent tweets search
        else:
            self._recent_tweets_search(
                keywords=keywords,
                twitter_credentials=twitter_credentials,
                start_time=start_time,
                end_time=end_time,
                bbox=bbox,
                country=country,
            )

    def _recent_tweets_search(
        self,
        start_time: datetime.datetime,
        end_time: datetime.datetime,
        twitter_credentials: list = None,
        keywords: list = None,
        bbox: str = None,
        country: str = None,
    ):
        twitter_api = credentials._get_twitter_api(
            db_conn=self._db_conn, purpose="search", credentials=twitter_credentials[0]
        )
        query = self._get_twitter_api_query(
            keywords=keywords, bbox=bbox, country=country
        )
        logger.info(f"Working on recent search with query {query}")

        since_id = SnowFlake.get_tweet_id_from_time(start_time)
        max_id = SnowFlake.get_tweet_id_from_time(end_time)
        logger.info(f"Using boundary tweet ids {since_id} and {max_id}")
        c = 0
        for page in tweepy.Cursor(
            twitter_api.search,
            q=query,
            count=100,
            since_id=since_id,
            max_id=max_id,
            tweet_mode="extended",
        ).pages():
            for status in page:
                extract.retrieve_content_from_tweet(
                    tweet=status._json,
                    db_conn=self._db_conn,
                    include_relations=self._include_relations,
                    include_users=self._include_users,
                )
                c += 1
        logger.info(f"Retrieved {c} tweets")

    def _full_archive_tweets_search(
        self,
        start_time: datetime.datetime,
        end_time: datetime.datetime,
        keywords: list = None,
        credentials: dict = None,
        onfly_sampling: bool = True,
        sampling_rate: int = 1,
        bbox: str = None,
        country: str = None,
    ):
        """Use full archive search feature of the v2 Academic API to recover tweets.

        Args:
            onfly_sampling: By default will retrieve all tweets specified in the given
                time period. If False, it will discretise the provided time range in
                seconds-wide blocks to sample 1% of all tweets using time-sampling.

        """

        # if no specified credentials and no env var loaded, query db
        credentials = self._get_academic_credentials(credentials=credentials)
        # here dump all retrieved tweets to output file to be later processed
        raw_output_filename, output_filename = self._get_output_filepaths(
            start_time=start_time, end_time=end_time, keywords_count=len(keywords)
        )

        if not onfly_sampling:
            return self._run_full_archive_continuous_search(
                raw_output_filename=raw_output_filename,
                output_filename=output_filename,
                start_time=start_time,
                end_time=end_time,
                keywords=keywords,
                credentials=credentials,
                bbox=bbox,
                country=country,
            )
        else:
            return self._run_full_archive_discrete_search(
                raw_output_filename=raw_output_filename,
                output_filename=output_filename,
                keywords=keywords,
                start_time=start_time,
                end_time=end_time,
                credentials=credentials,
                sampling_rate=sampling_rate,
                bbox=bbox,
                country=country,
            )

    def _run_full_archive_continuous_search(
        self,
        raw_output_filename: str,
        output_filename: str,
        credentials: dict,
        start_time: datetime,
        end_time: datetime,
        keywords: list = None,
        bbox: str = None,
        country: str = None,
        uid: str = None,
    ):
        """Directly search for all tweets posted in the given timeframes."""

        # retrieve all academic search args to search tweets
        search_args = self._get_academic_search_args(credentials)
        # generate query using searchtweets module
        search_query = self._get_academic_search_query(
            start_time, end_time, keywords=keywords, bbox=bbox, country=country, uid=uid
        )

        return self._execute_full_archive_search(
            search_query=search_query,
            search_args=search_args,
            raw_output_file=raw_output_filename,
            output_file=output_filename,
        )

    def _run_full_archive_discrete_search(
        self,
        raw_output_filename: str,
        output_filename: str,
        start_time: datetime.datetime,
        end_time: datetime.datetime,
        credentials: dict,
        keywords: str = None,
        use_tweet_ids: bool = False,
        sampling_rate: int = 1,
        bbox: str = None,
        country: str = None,
    ):
        """Discretize continuous time frame and run full archive search
        on each time block."""

        discrete_time_ranges = self._get_discrete_time_ranges(
            start_time=start_time,
            end_time=end_time,
            use_tweet_ids=use_tweet_ids,
            sampling_factor=sampling_rate,
        )

        # search all tweets for each given interval
        for discrete_time_range in discrete_time_ranges:
            self._run_full_archive_continuous_search(
                raw_output_filename=raw_output_filename,
                output_filename=output_filename,
                keywords=keywords,
                start_time=discrete_time_range[0],
                end_time=discrete_time_range[1],
                credentials=credentials,
                bbox=bbox,
                country=country,
            )

    def _get_discrete_time_ranges(
        self,
        start_time: datetime.datetime,
        end_time: datetime.datetime,
        sampling_factor: int = 1,
        sampling_frequency: int = 300,
        use_tweet_ids: bool = False,
    ):
        """Divide continuous time range into chunks.

        The goal of this sampling process is to imitate the Twitter Streaming
            API by randomly fully collecting 1% of the data over discrete
            time periods (at high frequency). Here, with a sampling factor of
            100, we could consider collecting 3 seconds of tweets every 5 minutes.
            Note that we randomly select the specific 3 seconds window within the
            5 minute time range (to avoid random synchronization with exogeneous
            event).

        Args:
            sampling_factor: Amplitude of the sample to be drawn (in percentage).
            sampling_frequency: Frequency of the sample to be drawn from Twitter API.

        """

        # get second size of sample
        sample_size = int(sampling_frequency * sampling_factor / 100)

        # get times in linux epoch ts, compute interval size
        start_ts = start_time.timestamp()
        end_ts = end_time.timestamp()
        sampling_space_size = int(end_ts - start_ts)

        # iterate over
        sampling_intervals = []
        for i in range(0, sampling_space_size, sampling_frequency):
            # determine second offest to get start_ts+i+offset, i.e. 0:0+300-3
            random_interval = random.randint(i, i + sampling_frequency - sample_size)
            tmp_start_ts = start_ts + random_interval
            tmp_end_ts = tmp_start_ts + sample_size

            tmp_start_time = datetime.datetime.fromtimestamp(tmp_start_ts)
            tmp_end_time = datetime.datetime.fromtimestamp(tmp_end_ts)

            if not use_tweet_ids:
                sampling_intervals.append((tmp_start_time, tmp_end_time))
            else:
                tmp_start_tid = SnowFlake.get_tweet_id_from_time(tmp_start_time)
                tmp_end_tid = SnowFlake.get_tweet_id_from_time(tmp_end_time)
                sampling_intervals.append((tmp_start_tid, tmp_end_tid))

        return sampling_intervals

    def _execute_full_archive_search(
        self,
        search_query,
        search_args,
        raw_output_file,
        output_file,
        max_iter=1,
        debug=False,
    ):
        """Execute full archive search using Twitter API V2 endpoint."""

        result_stream = searchtweets.ResultStream(
            request_parameters=search_query,
            max_tweets="no limit",
            **search_args,
        )

        iter_count = 0
        tweet_batch = []
        for response_tweet in result_stream.stream():
            if "next_token" in response_tweet or "newest_id" in response_tweet:
                if debug:
                    print(f"got {len(tweet_batch)} tweets")
                # for traceability, add token element
                tweet_batch.append(response_tweet)
                raw_response_doc = extract._get_raw_response_doc(
                    tweet_batch=tweet_batch
                )

                # load from dict response to Tweet & User objects
                loaded_tweets = convert.Converter.convert_v2_academic_restful_response_to_v1_standard(
                    response_batch=tweet_batch
                )
                # convert all objs to dicts
                loaded_tweet_dicts = [
                    loaded_tweet.to_dict() for loaded_tweet in loaded_tweets
                ]

                # extract & upload the converted tweets to issue db
                for loaded_tweet in loaded_tweet_dicts:
                    extracted_content = extract.retrieve_content_from_tweet(
                        loaded_tweet,
                        db_conn=self.db_conn,
                        include_users=self.include_users,
                        include_relations=self.include_relations,
                    )

                # also upload this full response to the db
                # NOTE: for testing purposes add timestamp as collection tags
                # this_collection_name = db_schema._get_collection_name_from_dt(
                #    start_time=json.loads(search_query)["start_time"],
                #    end_time=json.loads(search_query)["end_time"],
                # )
                self.db_conn.add_restful_response(
                    response=raw_response_doc,
                    # collection_tag=this_collection_name,
                    api_endpoint=versions.ApiVersion.ACADEMIC_V2,
                )

                if settings.STORE_TO_FILE:
                    # save raw tweets + objs
                    loader._save_to_jsonl(
                        content=tweet_batch, output_file=raw_output_file, mode="a"
                    )
                    loader._save_to_jsonl(
                        content=loaded_tweet_dicts, output_file=output_file, mode="a"
                    )

                tweet_batch = []
                iter_count += 1
            else:
                tweet_batch.append(response_tweet)

        return iter_count != 0

    def _get_output_filepaths(
        self, start_time, end_time, keywords_count=None, uid=None
    ):
        """Retrieve output filepaths to store raw responses, and locally filtered responses.
        Note that both files are jsonl files.
        """

        output_ts_str = "%mM%dD%YY_%Hh%Mm"
        start_time_str = datetime.datetime.strftime(start_time, output_ts_str)
        end_time_str = datetime.datetime.strftime(end_time, output_ts_str)
        file_format = ".jsonl"

        tmp_content = keywords_count if keywords_count else uid
        output_filename = f"full_archive_search_{start_time_str}_{end_time_str}_{tmp_content}_tweets{file_format}"
        output_filename = str(settings.OUTPUT_DIR.joinpath(output_filename))

        raw_output_filename = "raw_" + output_filename
        raw_output_filename = str(settings.OUTPUT_DIR.joinpath(raw_output_filename))

        return raw_output_filename, output_filename

    def _get_time_for_twitter_api(self, timestamp):
        """Manually convert the datetime object to the string format required by the Twitter API."""

        return timestamp.strftime(settings.TWITTER_API_V2_TIME_STR)

    def _get_academic_search_query(
        self, start_time, end_time, keywords=None, bbox=None, country=None, uid=None
    ):
        """Get searchtweets query for full archive search using the Academic V2 API."""

        searchtweets_str_format = "%Y-%m-%d %H:%M"
        searchtweets_query = self._get_twitter_api_query(
            keywords=keywords, bbox=bbox, country=country, uid=uid
        )

        # note the dum replace statement, buggy otherwise
        search_query = searching._get_request_parameters(
            query=searchtweets_query,
            start_time=self._get_time_for_twitter_api(start_time),
            end_time=self._get_time_for_twitter_api(end_time),
            tweet_fields=self._get_included_tweet_fields(),
            user_fields=self._get_included_user_fields(),
            expansions=self._get_searchtweets_expansions(),
        )

        return search_query

    def _get_academic_search_args(self, credentials):
        """Get searchtweets args for full archive search using the Academic V2 API."""

        full_archive_endpoint_url = "https://api.twitter.com/2/tweets/search/all"
        yaml_env_searchtweets_filename = ".searchtweets_env.yaml"
        yaml_env_dict = {
            "search_tweets_v2": {
                "endpoint": full_archive_endpoint_url,
                "bearer_token": credentials.bearer_token,
                "consumer_key": credentials.consumer_key,
                "consumer_secret": credentials.consumer_secret,
            }
        }
        yaml_env_filepath = self._create_yaml_env(
            filename=yaml_env_searchtweets_filename, env_dict=yaml_env_dict
        )

        # load credentials with twitter's searchtweets package
        # see https://github.com/twitterdev/search-tweets-python/tree/v2
        search_args = searchtweets.load_credentials(
            filename=str(yaml_env_filepath.absolute()),
            yaml_key="search_tweets_v2",
        )

        return search_args

    def _get_academic_credentials(self, credentials=None):
        """Load Academic API v2 credentials."""

        if not credentials and settings.ACADEMIC_API_BEARER_TOKEN is None:
            pass
        elif not credentials and settings.ACADEMIC_API_BEARER_TOKEN is not None:
            credentials = OmegaConf.create(
                {
                    "bearer_token": settings.ACADEMIC_API_BEARER_TOKEN,
                    "consumer_key": settings.ACADEMIC_API_CONSUMER_KEY,
                    "consumer_secret": settings.ACADEMIC_API_CONSUMER_SECRET,
                }
            )

        if credentials.bearer_token is None:
            raise ValueError(
                "Need to have academic api credentials for full archive search."
            )

        return credentials

    def _get_twitter_api_query(
        self,
        keywords: list = None,
        bbox: str = None,
        country: str = None,
        uid: str = None,
    ):
        """Simple query building to retrieve a tweet containing either of the specified
        keyword. Note that there is a difference between specifying the # character
        or not in front of each keyword."""

        query = ""
        if keywords:
            query += " OR ".join(keywords)
            query = "( " + query + " )"

        if bbox:
            query += f" bounding_box:[{bbox}]"

        if uid:
            query += f" from:{uid}"

        if country:
            # NOTE: requires access to a 2.0 GNIP api
            # query += f" profile_country:{country}"
            query += f" place_country:{country}"

        return query.strip()

    def _get_included_tweet_fields(self):
        """Retrieve list of tweet fields to be included in Twitter API responses. Goal
        is to retrieve tweets with the same amount of data as those returned using the
        tweepy streaming feature."""

        tweet_fields = [
            "id",
            "text",
            "attachments",
            "author_id",
            "context_annotations",
            "conversation_id",
            "created_at",
            "entities",
            "geo",
            "in_reply_to_user_id",
            "lang",
            "public_metrics",
            "possibly_sensitive",
            "referenced_tweets",
        ]

        return ",".join(tweet_fields)

    def _get_included_user_fields(self):
        """Retrieve list of user fields to be included in Twitter API responses."""

        user_fields = [
            "id",
            "name",
            "username",
            "created_at",
            "description",
            "entities",
            "location",
            "pinned_tweet_id",
            "profile_image_url",
            "protected",
            "public_metrics",
            "url",
            "verified",
        ]

        return ",".join(user_fields)

    def _get_searchtweets_expansions(self):
        """We need to reconstruct the hashtag, mention, retweet and quote relation
            originating from the published tweet. Thus, we request additional info
            about the user and referenced tweets in the retrieved tweet, to be able
            to process them 'normally' using the extract.retrieve_content_from_tweet
            method.

        Retrieved expansions:
            - author_id:    Retrieves the user object representing the Tweet's author.
            - referenced_tweets.id: Return tweet object referenced by this tweet.
            - in_reply_to_user_id:  Return a user object representing the user this
                requested tweet is a reply of.
            - entities.mentions.username: returns a user obkect for the user mentioned in
                the Tweet.
            - referenced_tweets.id.author_id: returns a user object for the author of the
                reference tweet.
            - geo.place_id: Full place object as mentioned in the original tweet.

        for more info see: https://developer.twitter.com/en/docs/twitter-api/expansions

        """

        expansions = [
            "author_id",
            "referenced_tweets.id",
            "in_reply_to_user_id",
            "attachments.media_keys",
            "attachments.poll_ids",
            "geo.place_id",
            "entities.mentions.username",
            "referenced_tweets.id.author_id",
        ]

        return ",".join(expansions)

    def search_tweets_with_keywords(
        self,
        keywords: list = None,
        search_field: str = "text",
        search_collection: str = "tweets",
        issue: str = None,
        output_issue: str = None,
    ):
        """Run filter search on database with a given set of keywords. Two methods are available:
        text search in aggregation (therefore directly using db computational resources) or
        iterative search (download tweets and iterate through the whole db locally). We draft a
        method for both and experiment.

        Note that a text index is required on the tweets collection in order to query it. Ran creation
            of text index on the 24 million tweets database (might be too much).
        """

        tmp_issue = self._get_issue(issue=issue)
        output_issue = tmp_issue if not output_issue else output_issue
        return searching.run_local_tweets_search(
            keywords=keywords,
            search_field=search_field,
            search_collection=search_collection,
            issue=tmp_issue,
            output_issue=output_issue,
            db_conn=self.db_conn,
        )

    def _create_yaml_env(self, filename, env_dict):
        """Transform dict to yaml env save in current root directory."""

        with open(filename, "w") as f:
            yaml.dump(env_dict, f)

        return settings.SNPIPELINE_HOME.joinpath(filename)

    def _is_valid_timestamp(self, timestamp):
        """Check if this timestamp goes back more than 7 days from current date."""

        rn_timestamp = datetime.datetime.now()
        seven_days_delta = datetime.timedelta(days=7)

        return (timestamp - rn_timestamp) <= seven_days_delta

    def fetch_users_timelines(
        self, usernames=None, uids=None, local_credentials=True, api_credentials=None
    ):
        # get a twitter api, optionally using the allowed credentials
        if local_credentials and not api_credentials:
            kwargs = dict(
                credentials=dict(
                    access_token=settings.ACCESS_TOKEN,
                    access_token_secret=settings.ACCESS_TOKEN_SECRET,
                    consumer_key=settings.CONSUMER_KEY,
                    consumer_secret=settings.CONSUMER_SECRET,
                )
            )
        elif api_credentials:
            assert (
                "access_token" in api_credentials
                and "access_token_secret" in api_credentials
                and "consumer_key" in api_credentials
                and "consumer_secret" in api_credentials
            )
            kwargs = dict(credentials=api_credentials)
        else:
            kwargs = dict(db_conn=self._db_conn)
        twitter_api = credentials._get_twitter_api(**kwargs)
        print(
            f"Retrived twitter api: {twitter_api} with local credentials ({local_credentials})"
        )

        if usernames:
            print("Starting to fetch timelines for {} usernames".format(len(usernames)))
            for idx, username in enumerate(usernames):
                history._get_user_history(
                    username=username,
                    twitter_api=twitter_api,
                    db_connection=self._db_conn,
                )
                if idx % 1000 == 0 and idx != 0:
                    print(f"Fetched {idx+1} timelines")

        elif uids:
            print("Starting to fetch timelines for {} uids".format(len(uids)))
            for idx, uid in enumerate(uids):
                history._get_user_history(
                    uid=uid, twitter_api=twitter_api, db_connection=self._db_conn
                )

                if idx % 1000 == 0 and idx != 0:
                    print(f"Fetched {idx+1} timelines")



from tweepipe import settings
from tweepipe.db import db_client, db_schema


class BaseClient:
    """
    Tweepipe data collection client.

    The tweepipe client wraps around the tweepy client to facilitate data storage
    to MongoDB.

    :param issue: Name of the ongoing issue, also the working database. Defaults to None.
    :type issue: str, optional
    :param include_users: Whether to extract user profiles upon processing the retrieved tweets.
        If so, the users will be stored in a separate collection. Defaults to True.
    :type include_users: bool, optional
    :param include_relations: Whether to extract relations (hashtags, quotes, mentions, retweets)
        upon processing. Like users, if extracted they will be stored under separate collections
        in the mongo database. Defaults to True.
    :type include_relations: bool, optional
    :param skip_db: Set to True to skip usage of a database, and not have persistent storage on
        collection tasks. Defaults to False.
    :type skip_db: bool, optional
    :param schema: Specify an index schema for the retrieved Twitter data. For example, an index
        on tweet ids can be useful to quickly get to a tweet. Defaults to db_schema.INDEX_V3.
    :type schema: dict, optional
    """

    def __init__(
        self,
        issue: str = None,
        include_users: bool = True,
        include_relations: bool = True,
        skip_db: bool = False,
        schema: dict = None,
    ):

        self._issue = issue
        self._include_users = include_users
        self._include_relations = include_relations
        self._schema = schema if schema else db_schema.INDEX_V3

        if skip_db:
            self._db_conn = None
        else:
            self._db_conn = db_client.DBClient(
                issue=issue,
                include_users=include_users,
                include_relations=include_relations,
                schema=schema,
            )

        self._env_file = None

    def _get_issue(self, issue: str = None) -> str:
        """Determine the current working issue.

        :param issue: specified working issue, defaults to None.
        :type issue: str, optional
        :return: Current working issue.
        :rtype: str
        """

        tmp_issue = issue if issue else self.issue

        return tmp_issue

    def _get_include_users(self, include_users: bool = False) -> bool:
        """
        Determine whether to include users, if either the client's flag is set,
        or this flag was set.


        :param include_users: include users flag, defaults to False.
        :type include_users: bool, optional
        :return: final include users flag
        :rtype: bool
        """

        return include_users or self.include_users

    def _get_include_relations(self, include_relations: bool = False) -> bool:
        """
        Determine whether to include relations, if either the client's flag is set,
        or this flag was set.

        :param include_relations: include relations flag, defaults to False.
        :type include_relations: bool, optional
        :return: final include relations flag
        :rtype: bool
        """

        return include_relations or self.include_relations

    def set_issue(self, new_issue: str):
        """Change name of working issue, and refresh db connection.

        :param new_issue: New issue name
        :type new_issue: str
        """

        if new_issue != self.issue:
            self._issue = new_issue
            self._db_conn = db_client.DBClient(
                issue=new_issue,
                include_users=self.include_users,
                include_relations=self.include_relations,
                schema=self.schema,
            )

    def set_mirror_collection_name(self, mirror_collection_name: str):
        """Set name of collection handling the storage of raw api response.

        :param mirror_collection_name: mirror collection name
        :type mirror_collection_name: str
        """

        self._db_conn.response_mirror_name = mirror_collection_name

    def set_include_users(self, include_users: bool = True):
        self._include_users = include_users

    def set_include_relations(self, include_relations: bool = True):
        self._include_relations = include_relations

    def set_env_file(self, env_file: str):
        """Update environment file path, and load it in settings.

        :param env_file: filepath to environment file
        :type env_file: str
        """

        self._env_file = env_file
        settings.load_config(env_file=self.env_file)

    @property
    def db_conn(self) -> db_client.DBClient:
        """Get current database connection.

        :return: Current client database connection.
        :rtype: db_client.DBClient
        """
        return self._db_conn

    @property
    def schema(self) -> dict:
        """Get current database index schema.

        :return: current working database index schema
        :rtype: dict
        """
        return self._schema

    @property
    def include_users(self) -> bool:
        """Get include users flag.

        :return: include users flag
        :rtype: bool
        """
        return self._include_users

    @property
    def include_relations(self) -> bool:
        """Get include relations flag.

        :return: include relations flag
        :rtype: bool
        """
        return self._include_relations

    @property
    def issue(self) -> str:
        """Get working subject, also working database name.

        :return: issue name
        :rtype: str
        """
        return self._issue

    @property
    def env_file(self) -> str:
        """Get environment file path.

        :return: env file path
        :rtype: str
        """
        return self._env_file



import pymongo
from pymongo import MongoClient
import bson
from typing import Any, Union
import time
import json

from loguru import logger

from tweepipe import settings
from tweepipe.db import db_schema
from tweepipe.utils import errors


class DBClient:
    """A simple and functional wrapper around the pymongo client to work with Twitter data
    in the JSON format. This client includes utilities to manage Twitter API credentials
    stored in a mongo database, as well as manage queues of data to be hydrated, fetched,
    or processed.

    :param issue: Name of the current working database, defaults to None.
    :type issue: str, optional
    :param response_mirror_name: Mirror collection to store raw API v2
        responses in addition to the tweets, defaults to None.
    :type response_mirror_name: str, optional
    :param include_relations: Whether to extract relations from tweets, defaults to True.
    :type include_relations: bool, optional
    :param include_users: Whether to extract relations from users, defaults to True.
    :type include_users: bool, optional
    :param batch_size: Batch size for bulk data lists, defaults to 128.
    :type batch_size: int, optional
    :param schema: Index schema for the working database, defaults to db_schema.DEFAULT_TWITTER_DB_SCHEMA.
    :type schema: dict, optional
    """

    def __init__(
        self,
        issue: str = None,
        response_mirror_name: str = None,
        include_relations: bool = True,
        include_users: bool = True,
        batch_size: int = 1024,
        schema: dict = None,
        declared_collections: list = None,
        block_index_create: bool = False,
    ):
        self.conn = MongoClient(
            settings.MONGO_HOST,
            settings.MONGO_PORT,
            username=settings.MONGO_USERNAME,
            password=settings.MONGO_PASSWORD,
            ssl=True,
        )
        self.block_index_create = block_index_create

        # determine db name
        self.declared_collections = declared_collections
        issue = self._get_db_name(db_name=issue)
        self.issue = issue if issue != None else "default"

        self.response_mirror_name = (
            response_mirror_name
            if response_mirror_name != None
            else settings.RESPONSE_MIRROR_COLLECTION_NAME
        )

        self.include_users = include_users
        self.include_relations = include_relations
        self.batch_size = batch_size
        self.bulk_data = {}

        self._db_schema = db_schema._get_db_schema(
            issue=issue,
            include_users=include_users,
            include_relations=include_relations,
            schema_base=schema if schema else db_schema.DEFAULT_TWITTER_DB_SCHEMA,
        )
        self.init_collections(_db_schema=self._db_schema)
        self.init_bulk_buffers()

    def init_bulk_buffers(self):
        """Create temporary buffers to push data in bulk to the database. Initialize
        buffers after the employed schema. If the user is trying to push data not
        available with the current schema, addition and push operations should fail.
        """

        if self.declared_collections:
            for collection_name in self.declared_collections:
                self.bulk_data[collection_name] = []

        for collection_name in self._db_schema.get(self.issue):
            self.bulk_data[collection_name] = []
        # logger.debug(f"Initialized bulk buffers for {', '.join(list(self.bulk_data))}")

    def init_collections(self, _db_schema: dict):
        """Re-create collections in MongoDB server using provided indexes."""

        for db_name in _db_schema:
            for collection_name in _db_schema[db_name]:
                # logger.info(f"Creating collection {collection_name} in db {db_name}")
                collection_conn = self.conn[db_name][collection_name]
                for index in _db_schema[db_name][collection_name]:
                    try:
                        if not self.block_index_create:
                            collection_conn.create_index(
                                index["index"],
                                unique=index.get("unique", False),
                            )
                    except pymongo.errors.OperationFailure as e:
                        pass

    def _get_collection(self, collection_name: str, db_name: str = None):
        """Retrieve connection to the specified database collection."""

        db_name = self.issue if not db_name else db_name

        return self.conn[db_name][collection_name]

    def clear_database(self, db_name: str = None, are_you_sure: bool = False):
        """Remove all collections and data from the specified database."""

        if not are_you_sure:
            raise ValueError("You must be sure before deleting a database!")

        db_name = self._get_db_name(db_name)

        if (
            self.conn.address != ("127.0.0.1", 27017)
            and self.conn.address
            != (
                "localhost",
                27017,
            )
            and not are_you_sure
        ):
            raise PermissionError(
                "Are you sure you want to delete a non-local database?"
            )

        try:
            self.conn.drop_database(db_name)
        except pymongo.errors.OperationFailure as e:
            pass

    def _load_bson_object_ids(self, batch: Union[list, set]) -> list:
        """Load string bson ids to Bson Python objects."""

        for doc in batch:
            doc["_id"] = bson.objectid.ObjectId(doc["_id"]["$oid"])

        return batch

    def _get_db_name(self, db_name: str = None):
        """Returns db_name if no settings specified, else the predefined name."""

        db_name = db_name if db_name is not None else settings.MONGO_DB

        return db_name

    def add_restful_response(
        self, response: list, collection_tag: str = None, api_endpoint: str = None
    ):
        """Add Academic/Standard api response to db."""

        # make sure restful response mirror of collection exists
        response_collection = self._check_create_response_mirror(
            collection_tag=collection_tag
        )

        # insert response document into collection
        response_collection.insert_one(response)

    def _check_create_response_mirror(
        self, collection_tag: str = None, api_endpoint: str = None
    ):
        """Checks whether or not raw response collections are available for
        the curent database. If not, init a new one.

        Args:
            collection_tag: Collection name to search for in the current db.

        Returns:
            Returns True if a new collection was created. Otherwise, returns False.
        """

        response_mirror_name = self._get_response_mirror_name(
            collection_tag=collection_tag, api_endpoint=api_endpoint
        )

        return self._get_collection(collection_name=response_mirror_name)

    def _get_response_mirror_name(
        self, collection_tag: str = None, api_endpoint: str = None
    ):
        """Checks for response mirror tag, otherwise retrieves default settings value."""

        collection_name = (
            collection_tag if collection_tag != None else self.response_mirror_name
        )
        if api_endpoint:
            collection_name += "".join(["_", api_endpoint])

        return collection_name

    def add_to_collection(self, collection_name: str, data: Any):
        """Add json data to its collection's buffer before pushing it to the database.

        :param collection_name: name of the collection to add data to
        :type collection_name: str
        :param data: data to be added
        :type data: Any
        :raises ValueError: only add data to collections available in working db
        """
        if collection_name not in self.bulk_data:
            raise ValueError("Undefined collection for current schema.")

        if isinstance(data, list):
            self.bulk_data[collection_name].extend(data)
        else:
            self.bulk_data[collection_name].append(data)

        if len(self.bulk_data[collection_name]) >= self.batch_size:
            self._push_bulk_data(collection_name)

    def add_tweet(self, tweet: dict):
        self.add_to_collection("tweets", tweet)

    def add_user(self, user: dict):
        self.add_to_collection("users", user)

    def add_like(self, like: dict):
        self.add_to_collection("likes", like)

    def add_media(self, media: Union[dict, list]):
        self.add_to_collection("media", media)

    def add_poll(self, poll: Union[dict, list]):
        self.add_to_collection("polls", poll)

    def add_place(self, place: Union[dict, list]):
        self.add_to_collection("places", place)

    def add_users(self, users: list):
        self.add_to_collection("users", users)

    def add_relation_extraction(self, relation_extraction: dict):
        self.add_to_collection("relation_extraction", relation_extraction)

    def add_fetching_uids(self, uids: list):
        self.add_to_collection("fetching_uids", uids)

    def add_missing_tids(self, tids: list):
        self.add_to_collection("missing_tids", tids)

    def add_hydrating_tids(self, tids: list):
        self.add_to_collection("hydrating_tids", tids)

    def add_missing_users(self, missing_uids: list):
        self.add_to_collection("missing_users", missing_uids)

    def add_hashtags(self, hashtags: list):
        self.add_to_collection("hashtags", hashtags)

    def add_retweets(self, retweets: list):
        self.add_to_collection("retweets", retweets)

    def add_mentions(self, mentions: list):
        self.add_to_collection("mentions", mentions)

    def add_quotes(self, quotes: list):
        self.add_to_collection("quotes", quotes)

    def add_replies(self, replies: list):
        self.add_to_collection("replies", replies)

    def add_scores(self, scores: list):
        self.add_to_collection("users", scores)

    def get_collection_paginator(
        self, collection_name: str, projection: dict = None, page_size: int = 2048
    ) -> pymongo.cursor.Cursor:

        collection = self._get_collection(collection_name)
        cursor = collection.find(
            projection=projection,
            batch_size=page_size,
            sort=[("_id", pymongo.ASCENDING)],
            allow_disk_use=True,
        )

        return cursor

    def create_hydrating_index(self):
        """Enable fast query on status field for hydrating tweets."""

        hydrating_tids_collection = self._get_collection(
            collection_name="hydrating_tids",
            db_name=self.issue,
        )
        hydrating_tids_collection.create_index(
            "status",
        )
        hydrating_tids_collection.create_index("tid", unique=True)

    def _push_bulk_data(self, collection_name: str):
        """Push a batch of data to the collection on the working database.

        :param collection_name: Collection name to push data to.
        :type collection_name: str
        """
        collection = self._get_collection(collection_name, db_name=self.issue)
        logger.info(f"Pushing to collection {collection_name} in db {self.issue}")
        try:
            logger.info(
                f"pushing {len(self.bulk_data[collection_name])} elements to {collection_name}"
            )
            if collection_name == "users":
                logger.info(
                    f"Individual user ids: {len(set([user.get('uid') for user in self.bulk_data[collection_name]]))}"
                )
                logger.info(
                    f"Number of records to insert: {len(self.bulk_data[collection_name])}"
                )
            # print(json.dumps(self.bulk_data[collection_name][0], indent=2))
            # time.sleep(100)
            collection.insert_many(self.bulk_data[collection_name], ordered=False)
        except pymongo.errors.BulkWriteError as e:
            print(f"Encountered error while pushing to {collection_name}")  # , e)
            pass

        # reset collection bulk buffer
        self.bulk_data[collection_name] = []

    def flush_content(self):
        """Flush cached tweets to db upon exit."""

        for collection_name in self.bulk_data:
            if len(self.bulk_data.get(collection_name)) > 0:
                self._push_bulk_data(collection_name)

    def reset_twitter_credentials_statuses(self):
        """Resets Twitter credentials use statuses."""

        api_collection = self._get_collection("api", db="api")

        return api_collection.update_many(
            filter={},
            update={"$set": {"lookup": 0, "search": 0, "stream": 0, "any": 0}},
        )

    def insert_twitter_credentials(self, twitter_credentials: list):
        """Insert Twitter API credentials into dataabse.

        Args:
            twitter_credentials: Dict-like object containing Twitter API
                credentials to add to the database. A Twitter API credential
                should contain the following fields: "access_token",
                "access_token_secret", "consumer_key", "consumer_secret".

        """

        api_collection = self._get_collection(collection_name="api", db_name="api")
        try:
            api_collection.insert_many(twitter_credentials, ordered=False)
        except pymongo.errors.BulkWriteError as e:
            pass

    def get(self, collection_name: str, params: dict = None):
        """List all documents in a collection."""
        collection = self._get_collection(collection_name, db_name=self.issue)
        return collection.find(filter=params)

    def get_twitter_credentials(self, count=1, purpose="any", must_be_free=False):
        """Retrieve Twitter API credentials.

        Args:
            credentials_count: Number of Twitter API credentials to retrieve.
            purpose: Use case motivating the retrieval. Useful for resource
                allocation.

        Returns:
            A list of retrieved apis.

        """

        api_collection = self._get_collection(collection_name="api", db_name="api")

        count = 0 if count == -1 else count
        params = dict(limit=count)

        # avoid having any busy api
        params["filter"] = {}
        if must_be_free:
            _filter = {purpose: 0}
            params["filter"].update(_filter)

        params["projection"] = dict(
            access_token=1, access_token_secret=1, consumer_key=1, consumer_secret=1
        )

        # retrieve free Twitter API credentials
        retrieved_credentials = list(api_collection.find(**params))

        # update their status in db
        bulk_updates = []
        for credential in retrieved_credentials:
            bulk_updates.append(
                pymongo.UpdateOne(
                    {"access_token": credential["access_token"]}, {"$set": {purpose: 1}}
                )
            )
        if bulk_updates:
            api_collection.bulk_write(bulk_updates)

        # raise an error if could not retrieve the right amount of credentials
        if len(retrieved_credentials) < count:
            raise errors.SnPipelineError(
                errors.SnPipelineErrorMsg.UNSUFFICIENT_FREE_RESOURCES
            )

        return retrieved_credentials

    def update_twitter_credentials(
        self, consumer_key: str, purpose: str, status: int = 0
    ):
        """Update the consumer key utilization status."""

        query = {"consumer_key": consumer_key}
        update = {"$set": {f"{purpose}": status}}
        api_collection = self._get_collection(collection_name="api", db_name="api")
        api_collection.update_one(query, update)



from datetime import datetime


class SnowFlake:
    """
    Reverse engineering of the tweet id generation mechanism, based
    on current timestamp and integer offset.

    :return: defaults to None.
    :rtype: None.
    """

    @classmethod
    def get_timestamp_from_tweet_id(cls, tweet_id: int):
        """Get tweet id timestamp.

        :param tweet_id: Tweet id to reverse engineer.
        :type tweet_id: int
        :return: Integer timestamp corresponding to tweet id.
        :rtype: int
        """

        offset = 1288834974657
        tstamp = (tweet_id >> 22) + offset

        return tstamp

    @classmethod
    def get_tweet_id_from_time(cls, time_obj: datetime):
        """Convert time specs into a tweet id.

        :param time_obj: Datetime object to convert to tweet id.
        :type time_obj: datetime
        :return: Inferred tweet id.
        :rtype: int
        """

        offset = 1288834974657
        ts = int(datetime.timestamp(time_obj) * 1000)
        ts -= offset
        tweet_id = ts << 22

        return tweet_id

    @classmethod
    def get_datetime_from_tweet_id(cls, tweet_id: int):
        """Retrieve datetime object from tweet id.

        :param time_obj: Tweet id to reverse engineer.
        :type time_obj: int
        :return: Inferred datetime from tweet id.
        :rtype: datetime
        """

        ts = cls.get_timestamp_from_tweet_id(tweet_id)
        utcdttime = datetime.utcfromtimestamp(ts / 1000)

        return utcdttime



import json
from collections import namedtuple

import tweepy
from loguru import logger
from tweepy import StreamRule
from tweepy.client import Response
from tweepy.tweet import Tweet

from tweepipe.db import db_client, db_schema
from tweepipe.legacy.utils import extract

StreamResponse = namedtuple(
    "StreamResponse", ("data", "includes", "errors", "matching_rules")
)


class StreamV1(tweepy.Stream):
    def __init__(
        self,
        db_conn: db_client.DBClient,
        output_file: str,
        consumer_key: str,
        consumer_secret: str,
        access_token: str,
        access_token_secret: str,
        **kwargs,
    ):
        """
        Streaming client for the Twitter API v1.

        Args:
            db_conn (db_client.DBClient): Active db connection.
            output_file (str): Output file for data not stored to db.
            consumer_key (str): v1 API consumer key.
            consumer_secret (str): v1 API consumer secret.
            access_token (str): v1 API access token.
            access_token_secret (str): v1 API access token secret.
        """
        super().__init__(
            consumer_key, consumer_secret, access_token, access_token_secret, **kwargs
        )

        self.db_conn = db_conn
        self.output_file = output_file
        self.consumer_key = consumer_key
        self.consumer_secret = consumer_secret
        self.access_token = access_token
        self.access_token_secret = access_token_secret

    def on_data(self, raw_data: str) -> bool:
        """
        Process incoming data for API v1.

        Args:
            raw_data (str): Raw data packet

        Returns:
            bool: Whether the data was correctly processed.
        """
        # shady way of extracting information from the tweet
        data = json.loads(raw_data)
        if "id_str" not in data:
            return True

        extract.retrieve_content_from_tweet(
            data,
            db_conn=self.db_conn,
            output_file=self.output_file,
            include_users=self.db_conn.include_users,
            include_relations=self.db_conn.include_relations,
        )

        return True


class StreamingClient(tweepy.StreamingClient):
    def __init__(
        self,
        bearer_token: str,
        issue: str = None,
        db_conn: db_client.DBClient = None,
        include_users: bool = False,
        include_relations: bool = False,
        return_type: type = Response,
        wait_on_rate_limit: bool = False,
        **kwargs,
    ):
        """
        _summary_

        Args:
            bearer_token (str): Twitter API v2 bearer token.
            issue (str, optional): Name of issue. Defaults to None.
            db_conn (db_client.DBClient, optional): Active db connection. Defaults to None.
            include_users (bool, optional): Whether to parse users on
                the fly. Defaults to False.
            include_relations (bool, optional): Whether to parse relations
                on the fly. Defaults to False.
            return_type (type, optional): Format of returned response from the Twitter API.
                Defaults to Response.
            wait_on_rate_limit (bool, optional): Wait when quota reached. Defaults to False.
        """
        super(tweepy.StreamingClient, self).__init__(
            bearer_token,
            return_type=return_type,
            wait_on_rate_limit=wait_on_rate_limit,
            **kwargs,
        )

        # optionally create a new database connection
        self.issue = issue
        if not db_conn:
            self.db_conn = db_client.DBClient(
                issue=issue,
                include_relations=include_relations,
                include_users=include_users,
                schema=db_schema.DEFAULT_ACADEMIC_V2_SCHEMA,
            )
        else:
            self.db_conn = db_conn

    def on_tweet(self, tweet: StreamResponse):
        """
        Store tweet to database.

        Args:
            tweet (StreamResponse): Received tweet to be stored.
        """
        self.db_conn.add_tweet(tweet.data)

    def on_data(self, raw_data: str):
        """
        Process incoming API packet.

        Args:
            raw_data (str): Raw data to be processed.
        """
        # overriding definition of on_data
        data = json.loads(raw_data)

        tweet = None
        includes = {}
        errors = []
        matching_rules = []

        if "data" in data:
            tweet = Tweet(data["data"])
            self.on_tweet(tweet)
        if "includes" in data:
            self.on_includes(includes)
        if "errors" in data:
            errors = data["errors"]
            self.on_errors(errors)
        if "matching_rules" in data:
            matching_rules = [
                StreamRule(id=rule["id"], tag=rule["tag"])
                for rule in data["matching_rules"]
            ]
            self.on_matching_rules(matching_rules)

        self.on_response(StreamResponse(tweet, includes, errors, matching_rules))

    def on_includes(self, includes: dict):
        """
        Add included attachments to db.

        Args:
            includes (dict): All attachment included in received API packet.

        Raises:
            ValueError: Unknown attachment type.
        """
        for key in includes:
            if key == "users":
                self.db_conn.add_users(includes[key])
                logger.info(
                    f"Adding {len(includes[key])} more users to the {self.issue} database."
                )
            elif key == "polls":
                self.db_conn.add_poll(includes[key])
            elif key == "media":
                self.db_conn.add_media(includes[key])
            elif key == "places":
                self.db_conn.add_place(includes[key])
            elif key == "tweets":
                self.db_conn.add_tweets(includes[key])
            else:
                raise ValueError("Unknown include key: {}".format(key))



import time
import math
from concurrent import futures
from loguru import logger
from typing import Any

import tweepy

from tweepipe.utils import tracker


def run_parallel(fn: Any, kwargs_list: list, max_workers: int = None) -> list:
    """Parallelize a function call.

    :param fn: Function to be run in parallel.
    :type fn: function
    :param kwargs_list: List of keyword arguments to be provided with
        function call.
    :type kwargs_list: list
    :param max_workers: Maximum number of process employed to run function,
        defaults to None.
    :type max_workers: int, optional
    :return: Results returned by individual function calls.
    :rtype: list
    """

    executor_kwargs = dict(max_workers=max_workers)
    as_completed_fn = futures.as_completed

    with futures.ProcessPoolExecutor(**executor_kwargs) as executor:
        # log start of streaming/searching process
        for kwargs in kwargs_list:
            if "keywords" in kwargs:
                logger.info(
                    f'Running streaming for {kwargs["issue"]} with keywords: {kwargs["keywords"]}.'
                )

        # initial submission of the streaming tasks
        futures_list = {executor.submit(fn, **kwargs): kwargs for kwargs in kwargs_list}

        results = []
        for i, future in enumerate(as_completed_fn(futures_list)):
            kwargs = futures_list[future]
            try:
                result = future.result()
            except Exception as e:
                error_msg = f"Error while running parallel task with args {kwargs}."
                tracker.log_error(e, msg=error_msg)
                logger.error(error_msg)

            results.append(result)
            logger.info(f"Execution done for ({i+1}/{len(kwargs_list)})")

    return results


def get_lookup_users_kwargs_list(
    uids: list = [], apis: list = [], issue: str = None, env_file: str = None
):
    uid_batch_size = math.ceil(len(uids) / len(apis))
    kwargs_list = []

    for i, api in zip(list(range(0, len(uids), uid_batch_size)), apis):
        kwargs_list.append(
            {
                "uids": uids[i : i + uid_batch_size],
                "twitter_api": api,
                "issue": issue,
                "env_file": env_file,
            }
        )

    return kwargs_list



import argparse
import os

import tweepy
import twint
from loguru import logger

from tweepipe import settings
from tweepipe.base import activity
from tweepipe.db import db_client
from tweepipe.utils import credentials, loader, parallel


def _fetch_followers_twint(
    uid_batch,
    output_folder: str = "./followers",
):
    """Unit method fetching followers ids for provided twitter handles batch
    using the Twint project toolkit."""

    for uid in uid_batch:
        tmp_conf = twint.Config()
        tmp_conf.Username = uid
        tmp_conf.Store_object = True

        twint.run.Followers(tmp_conf)
        followers = twint.output.follows_list

        # log user on current progress
        logger.info(f"Completed fetching of {len(followers)} followers for {uid}.")

        # save followers to json file
        output_file = f"{uid}.json"
        loader._save_to_json(followers, os.path.join(output_folder, output_file))


def _fetch_followers(
    uid_batch,
    twitter_api,
    output_folder: str = "./followers",
    verbose: bool = True,
    cap_followers: int = None,
    log_capping_file: str = "capped_users.txt",
    error_file: str = "error_users.txt",
    followers: bool = True,
    friends: bool = False,
):
    """Unit method fetching followers ids for provided twitter handles batch."""

    log_capping_file_path = os.path.join(output_folder, log_capping_file)
    error_file_path = os.path.join(output_folder, error_file)

    if followers:
        api_func = twitter_api.followers_ids
    elif friends:
        api_func = twitter_api.friends_ids
    else:
        raise ValueError("Must specify either followers or friends.")

    logger.info(f"Fetching friends for {len(uid_batch)} uids.")

    for uid in uid_batch:
        follower_cursor = tweepy.Cursor(api_func, id=uid)
        followers = []
        try:
            for page in follower_cursor.pages():
                followers.extend(page)

                if verbose and len(followers) != 0 and len(followers) % 50000 == 0:
                    logger.info(
                        f"Completed fetching of {len(followers)}/X followers for {uid}."
                    )

                # optionally cap number of followers to retrieve
                if cap_followers and len(followers) >= cap_followers:
                    logger.info(
                        f"Capping followers fetching for {uid} at {len(followers)}."
                    )
                    with open(log_capping_file_path, "a") as f:
                        f.write(str((uid, len(followers))) + "\n")
                    break
        except tweepy.error.TweepError as e:
            logger.info(f"Error while fetching {uid}.")
            with open(error_file_path, "a") as f:
                f.write(str(uid) + "\n")

        # log user on current progress
        logger.info(f"Completed fetching of {len(followers)} followers for {uid}.")

        # save followers to json file
        output_file = f"{uid}.json"
        loader._save_to_json(followers, os.path.join(output_folder, output_file))


def get_followers_parallel(
    uids,
    api_count,
    output_folder,
    twitter_credentials,
    cap_followers=None,
    batch_size=2048,
    followers=True,
    friends=False,
):
    """Organise sequential scraping for"""

    # num apis must be at least nimber of uids
    api_count = min(api_count, len(uids))
    logger.info(f"Retrieving {api_count} twitter api accesses.")

    twitter_apis, twitter_credentials = credentials._get_twitter_apis(
        twitter_credentials=twitter_credentials
    )
    print("got", len(twitter_apis), "twitter apis")
    max_workers = len(twitter_apis)
    batch_size = int(len(uids) / max_workers)

    # batch the uids to retrieve followers for
    uid_batches = loader._get_batches(uids, batch_size=batch_size)
    process_batches = activity._get_process_batches(uid_batches, twitter_apis)
    logger.info(
        f"Fetching followers for {len(uids)} uids in {len(uid_batches)} batches with {max_workers} processes."
    )

    _ = parallel.run_parallel(
        **{
            "fn": _fetch_followers,
            "kwargs_list": [
                {
                    "uid_batch": batch["uid_batch"],
                    "twitter_api": batch["twitter_api"],
                    "output_folder": output_folder,
                    "cap_followers": cap_followers,
                    "followers": followers,
                    "friends": friends,
                }
                for batch in process_batches
            ],
            "max_workers": max_workers,
        }
    )


def get_followers_sequential(
    uids,
    cap_followers,
    output_folder,
    twitter_credentials=None,
    backend="tweepy",
    followers=True,
    friends=False,
):
    """Sequentially fetch followers ids for each specified user id/screen name."""

    if not twitter_credentials:
        raise ValueError("Need twitter credentials")
    else:
        twitter_api = credentials._get_twitter_api(credentials=twitter_credentials)

    # exectue
    if backend == "tweepy":
        _fetch_followers(
            uids,
            twitter_api,
            output_folder,
            cap_followers=cap_followers,
            followers=followers,
            friends=friends,
        )
    elif backend == "twint":
        _fetch_followers_twint(uids, output_folder)
    else:
        raise ValueError


def get_followers_from_file(
    file,
    api_count: int = 1,
    twitter_credentials: list = [],
    env_file: str = None,
    parallel: bool = False,
    output_folder: str = "./followers",
    cap_followers: int = None,
    followers: bool = True,
    friends: bool = False,
):
    """Centralise followers ids retrieving process (dispatch in sequential/parallel process."""

    # use as many APIs as possible
    os.makedirs(output_folder, exist_ok=True)
    settings.load_config(env_file)

    uids = loader._load_from_file(file)
    logger.info("Loaded {} uids from file.".format(len(uids)))
    logger.info(f"Capping followers at {cap_followers}.")

    if parallel or api_count != 1 or len(twitter_credentials) > 0:
        get_followers_parallel(
            uids=uids,
            twitter_credentials=twitter_credentials,
            api_count=api_count,
            output_folder=output_folder,
            cap_followers=cap_followers,
            followers=followers,
            friends=friends,
        )
    else:
        get_followers_sequential(
            uids=uids,
            twitter_credentials=settings.TWITTER_CREDENTIALS_V1,
            output_folder=output_folder,
            cap_followers=cap_followers,
            followers=followers,
            friends=friends,
        )


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Load user(s) followers.")
    parser.add_argument(
        "--file",
        type=str,
        help="File containing user ids to fetch followers for.",
        required=True,
    )
    parser.add_argument(
        "--output-folder",
        type=str,
        help="Output file.",
        required=False,
        default="./followers",
    )
    parser.add_argument(
        "--api-count",
        type=int,
        help="Number of APIs to use in the process.",
        required=False,
        default=1,
    )
    parser.add_argument(
        "--cap-followers",
        type=int,
        help="Max number of followers to retrieve.",
        required=False,
        default=1000000,
    )
    parser.add_argument(
        "--env-file",
        type=str,
        help="File containing environment variables.",
        required=False,
        default=None,
    )
    parser.add_argument(
        "--parallel",
        type=str,
        help="Extract user profiles from streamed content.",
        required=False,
        default="no",
    )
    parser.add_argument(
        "--followers",
        type=str,
        help="Retrieve users' followers.",
        required=False,
        default="yes",
    )
    parser.add_argument(
        "--friends",
        type=str,
        help="Retrieve users' friends.",
        required=False,
        default="no",
    )

    args = parser.parse_args()

    file = args.file
    api_count = args.api_count
    parallel = args.parallel != "no"
    output_folder = args.output_folder
    env_file = args.env_file
    followers = args.followers != "no"
    friends = args.friends != "no"
    cap_followers = args.cap_followers

    get_followers_from_file(
        file=file,
        env_file=env_file,
        api_count=api_count,
        parallel=parallel,
        output_folder=output_folder,
        cap_followers=cap_followers,
        followers=followers,
        friends=friends,
    )



import json
import datetime

import pymongo
from tweepipe.legacy.utils import extract
import tweepy
from loguru import logger

from tweepipe import settings
from tweepipe.db import db_client, db_schema
from tweepipe.utils import parallel


def run_local_tweets_search(
    keywords: list,
    search_field: str,
    search_collection: str,
    output_issue: str,
    issue: str,
    db_conn: db_client.DBClient,
    env_file: str = None,
):
    """We search for all tweets which may contain a relevant keyword. Note that we strive
    to gather as much data as possible. If a retweet contains relevant keywords then it
    should be included here.
    """
    # assemble all user ids for which tweets have been collected
    fetched_uids = _get_uids_fetched(issue=issue, db_conn=db_conn)
    kwargs_list = _get_local_tweets_search_kwargs(
        fetched_uids=fetched_uids,
        keywords=keywords,
        search_field=search_field,
        search_collection=search_collection,
        output_issue=output_issue,
        issue=issue,
        env_file=env_file,
    )

    results = parallel.run_parallel(
        fn=_run_local_tweets_search_exec,
        kwargs_list=kwargs_list,
        max_workers=settings.WORKER_COUNT,
    )

    # sum up results of processing
    added_tweet_count, processed_tweet_count = 0, 0
    for result in results:
        added_tweet_count += result[0]
        processed_tweet_count += result[1]

    logger.info(
        f"Successfully filtered {added_tweet_count} tweets with {len(keywords)} keywords ({float(100*added_tweet_count/processed_tweet_count):.2f}%)."
    )

    return added_tweet_count, processed_tweet_count


def _run_local_tweets_search_exec(
    uids: list,
    search_collection: str,
    search_field: str,
    issue: str,
    output_issue: str,
    keywords: list,
    env_file: str = None,
    batch_size: int = 1024,
):
    if env_file:
        settings.load_config(env_file=env_file)

    input_conn = db_client.DBClient(issue=issue)
    input_collection = input_conn._get_collection(search_collection, db_name=issue)
    output_conn = db_client.DBClient(
        issue=output_issue,
        schema=db_schema.INDEX_V3,
        include_relations=True,
        include_users=True,
    )

    # re-extract data for all uids
    processed_doc_count = 0
    added_doc_count = 0
    for idx, uid in enumerate(uids):
        db_query = {"uid": uid}
        cursor = input_collection.find(
            filter=db_query,
            batch_size=batch_size,
            sort=[("uid", pymongo.ASCENDING)],
            allow_disk_use=True,
        )
        for doc in cursor:
            if is_tweet_related(
                doc["json"], keywords=keywords, search_field=search_field
            ):
                processed_tweet = extract.retrieve_content_from_tweet(
                    doc["json"],
                    include_users=output_conn.include_users,
                    include_relations=output_conn.include_relations,
                    db_conn=output_conn,
                )
                added_doc_count += 1
            processed_doc_count += 1

        if (idx + 1) % 64 == 0:
            logger.info(
                f"Finished processing {(idx+1)} users. Found {added_doc_count} related tweets ({float(100*added_doc_count/processed_doc_count):.2f}%)."
            )

    return added_doc_count, processed_doc_count


def is_tweet_related(tweet: dict, keywords: list, search_field: str = "text"):
    """Check if tweet is related to explored content."""

    def is_related(text, keywords):
        """Search subtext contained in tweet for a mentioned keyword."""
        text = text.lower().strip().replace("ー", "-")
        for keyword in keywords:
            if keyword in text:
                return True
        return False

    # produce set of keywords in lowercase for searching
    str_strip = lambda x: x.strip().lower().replace("ー", "-")
    keywords = list(map(str_strip, keywords))

    if "quoted_status" in tweet and is_related(
        tweet["quoted_status"][search_field], keywords
    ):
        return True
    if (
        "retweeted_status" in tweet
        and "quoted_status" in tweet["retweeted_status"]
        and is_related(
            tweet["retweeted_status"]["quoted_status"][search_field], keywords
        )
    ):
        return True
    if "retweeted_status" in tweet and is_related(
        tweet["retweeted_status"][search_field], keywords
    ):
        return True
    if is_related(tweet[search_field], keywords):
        return True
    else:
        return False


def _get_uids_fetched(issue: str, db_conn: db_client.DBClient):
    query = {"status": 2}
    collection = db_conn._get_collection("fetching_uids", db_name=issue)
    fetched_uid_docs = collection.find(
        filter=query, sort=[("tid", pymongo.ASCENDING)], allow_disk_use=True
    )
    fetched_uids = [doc["uid"] for doc in fetched_uid_docs]

    return fetched_uids


def _get_local_tweets_search_kwargs(
    fetched_uids: str,
    keywords: list,
    search_field: str,
    search_collection: str,
    output_issue: str,
    issue: str,
    env_file: str = None,
):
    uid_batch_size = int(len(fetched_uids) / settings.WORKER_COUNT) + 1
    uid_batches = [
        fetched_uids[i : i + uid_batch_size]
        for i in range(0, len(fetched_uids), uid_batch_size)
    ]
    kwargs_list = [
        {
            "uids": uid_batch,
            "search_collection": search_collection,
            "search_field": search_field,
            "output_issue": output_issue,
            "issue": issue,
            "keywords": keywords,
            "env_file": env_file,
        }
        for uid_batch in uid_batches
    ]

    return kwargs_list


def _get_request_parameters(
    query: str,
    start_time: str = None,
    end_time: str = None,
    since_id: str = None,
    until_id: str = None,
    tweet_fields: str = None,
    user_fields: str = None,
    expansions: str = None,
    results_per_call: int = 100,
):
    """Re-implementing searchtweets' gen_request_parameters function for lack
    of handling of second timestamps."""

    payload = {"query": query}
    # number of results to be returned
    payload["max_results"] = results_per_call

    if start_time:
        payload["start_time"] = start_time
    if end_time:
        payload["end_time"] = end_time
    if since_id:
        payload["since_id"] = since_id
    if until_id:
        payload["until_id"] = until_id
    if tweet_fields:
        payload["tweet.fields"] = tweet_fields
    if user_fields:
        payload["user.fields"] = user_fields
    if expansions:
        payload["expansions"] = expansions

    return json.dumps(payload)


def _get_fetch_from_db_histories_kwargs(
    twitter_apis: list,
    since_ts: datetime.datetime,
    env_file: str,
    issue: str,
    include_users: bool,
    include_relations: bool,
    batch_size: int = 1024,
):
    keys = list(range(1, len(twitter_apis) + 1))
    kwargs_list = []
    for api, key in zip(twitter_apis, keys):
        kwargs = {
            "twitter_api": api,
            "include_users": include_users,
            "include_relations": include_relations,
            "issue": issue,
            "env_file": env_file,
            "since_ts": since_ts,
            "batch_size": batch_size,
        }
        kwargs_list.append(kwargs)

    return kwargs_list


def _fetch_history_from_db(
    twitter_api,
    include_users,
    include_relations,
    issue,
    env_file,
    since_ts,
    batch_size,
):
    # get db connection on source & sink database
    db_conn = db_client.DBClient(
        issue=issue,
        include_relations=include_relations,
        include_users=include_users,
        schema=db_schema.INDEX_V3,
    )

    # iterate for tweet ids and lookup
    valid_uid_updates, invalid_uid_updates = [], []
    for uid in uids:
        tmp_cursor = tweepy.Cursor(
            twitter_api.user_timeline,
            id=uid,
            include_rts=True,
            count=200,
            exclude_replies=False,
            tweet_mode="extended",
        )
        try:
            iterate_history_cursor(db_conn, tmp_cursor, since_ts)
            valid_uid_updates.append(
                pymongo.operations.UpdateOne({"uid": uid}, {"$set": {"status": 2}})
            )
        except tweepy.TweepError as e:
            invalid_uid_updates.append(
                pymongo.operations.UpdateOne({"uid": uid}, {"$set": {"status": -1}})
            )

    if len(invalid_uid_updates) > 0:
        try:
            fetching_uids_collection = db_conn._get_collection(
                "fetching_uids", db_name=issue
            )
            fetching_uids_collection.bulk_write(invalid_uid_updates, ordered=False)
        except pymongo.errors.BulkWriteError as e:
            pass

    if len(valid_uid_updates) > 0:
        try:
            fetching_uids_collection = db_conn._get_collection(
                "fetching_uids", db_name=issue
            )
            fetching_uids_collection.bulk_write(valid_uid_updates, ordered=False)
        except pymongo.errors.BulkWriteError as e:
            pass


def _get_fetch_histories_kwargs(
    twitter_apis,
    twitter_credentials,
    uids,
    since_ts,
    issue,
    include_users,
    include_relations,
):
    """Produce batch of users to lookup since a given date."""

    batch_size = int(len(uids) / len(twitter_apis)) + 1
    kwargs_list = []
    for idx, i in enumerate(range(0, len(uids), batch_size)):
        uid_batch = uids[i : i + batch_size]
        twitter_api = twitter_apis[idx % len(twitter_apis)]
        twitter_credential = twitter_credentials[idx % len(twitter_apis)]
        kwargs = {
            "twitter_api": twitter_api,
            "twitter_credential": twitter_credential,
            "uids": uid_batch,
            "include_users": include_users,
            "include_relations": include_relations,
            "issue": issue,
            "since_ts": since_ts,
        }
        kwargs_list.append(kwargs)

    return kwargs_list


def _fetch_history(
    twitter_api: tweepy.API,
    twitter_credential: dict,
    uids: list,
    since_ts: datetime.datetime = None,
    issue: str = None,
    include_users: bool = True,
    include_relations: bool = True,
):
    """Lookup user histories until a given date."""

    # get db client
    db_conn = db_client.DBClient(
        issue=issue,
        include_relations=include_relations,
        include_users=include_users,
        schema=db_schema.INDEX_V3,
    )

    # push uids to db for tracking
    processed_uid_docs = [{"uid": uid, "status": 0} for uid in uids]
    db_conn.add_fetching_uids(processed_uid_docs)
    if len(db_conn.fetching_uids_batch) > 0:
        db_conn._push_fetching_uids_to_db()

    # iterate for tweet ids and lookup
    valid_uid_updates, invalid_uid_updates = [], []
    for uid in uids:
        tmp_cursor = tweepy.Cursor(
            twitter_api.user_timeline,
            id=uid,
            include_rts=True,
            count=200,
            exclude_replies=False,
            tweet_mode="extended",
        )
        try:
            iterate_history_cursor(db_conn, tmp_cursor, since_ts)
            valid_uid_updates.append(
                pymongo.operations.UpdateOne({"uid": uid}, {"$set": {"status": 2}})
            )
        except tweepy.TweepError as e:
            invalid_uid_updates.append(
                pymongo.operations.UpdateOne({"uid": uid}, {"$set": {"status": -1}})
            )

    if len(invalid_uid_updates) > 0:
        try:
            fetching_uids_collection = db_conn._get_collection(
                "fetching_uids", db_name=issue
            )
            fetching_uids_collection.bulk_write(invalid_uid_updates, ordered=False)
        except pymongo.errors.BulkWriteError as e:
            pass

    if len(valid_uid_updates) > 0:
        try:
            fetching_uids_collection = db_conn._get_collection(
                "fetching_uids", db_name=issue
            )
            fetching_uids_collection.bulk_write(valid_uid_updates, ordered=False)
        except pymongo.errors.BulkWriteError as e:
            pass


def iterate_history_cursor(
    db_conn: db_client.DBClient, cursor: tweepy.Cursor, since_ts: datetime.datetime
):
    """Iterate the history cursor to retrieve all user's tweets."""
    for page in cursor.pages():
        for raw_tweet in page:
            tweet_postdate = extract._get_creation_time_stamp(
                raw_tweet._json["created_at"]
            )
            if tweet_postdate > since_ts:
                extract.retrieve_content_from_tweet(
                    raw_tweet._json,
                    db_conn=db_conn,
                    include_users=db_conn.include_users,
                    include_relations=db_conn.include_relations,
                )
            else:
                return
    return



import argparse
import json
import os
import pickle
import time
from datetime import datetime
from queue import Queue

import tweepy
from loguru import logger

from tweepipe.base import activity
from tweepipe.db import db_client
from tweepipe.legacy.utils import include
from tweepipe.utils import credentials, loader, parallel


def _get_since_timestamp(since_str):
    return int(datetime.strptime(since_str, "%Y-%m-%d").timestamp())


def _get_tweet_timestamp(created_at_str):
    return int(include._get_creation_time_stamp(created_at_str).timestamp())


def _get_user_history(
    uid=None, username=None, twitter_api=None, db_conn=None, since_ts=0
):
    kwargs = dict(include_rts=True, count=200, exclude_replies=False)
    if uid:
        kwargs["user_id"] = uid
    elif username:
        kwargs["screen_name"] = username
    else:
        raise ValueError("Either uid or username must be provided.")
    cursor = tweepy.Cursor(twitter_api.user_timeline, **kwargs)
    tweet_count = 0

    try:
        for page in cursor.pages():
            for raw_tweet in page:
                _ = include.retrieve_content_from_tweet(
                    raw_tweet._json,
                    include_users=db_conn.include_users,
                    include_relations=db_conn.include_relations,
                    db_conn=db_conn,
                )
                tweet_count += 1
    except tweepy.errors.Unauthorized:
        print(f"Unauthorized error - user {uid if uid else username}")
    except tweepy.errors.NotFound:
        print(f"User not found - user {uid if uid else username}")
    except Exception as e:
        print("Error:", e, "user:", uid if uid else username)

    return tweet_count


def _get_batch_histories(
    uid_batch,
    twitter_api,
    since="2020-08-01",
    issue="history",
    db_conn=None,
    include_relations=False,
    include_users=False,
    output_folder="./",
):
    since_ts = _get_since_timestamp(since)

    if not db_conn:
        db_conn = db_client.DBClient(
            issue=issue,
            include_relations=include_relations,
            include_users=include_users,
        )
    missing_uids = []

    for uid in uid_batch:
        try:
            # by default retrieve as much information as possible
            fetched_tweets = _get_user_history(
                uid, twitter_api, db_conn, since_ts=since_ts
            )
            logger.debug(f"Fetched {len(fetched_tweets)} tweets for {uid}.")

            output_file_path = os.path.join(
                output_folder, "tweets_" + str(uid) + ".json"
            )

            # filter out mongo oid and convert datetime to str
            for tweet in fetched_tweets:
                for u in tweet["users"]:
                    if "_id" in u:
                        del u["_id"]

                for q in tweet["relations"]["quote"]:
                    q["created_at"] = q["created_at"].strftime("%m/%d/%Y, %H:%M:%S")
                    if "_id" in q:
                        del q["_id"]

                for h in tweet["relations"]["hashtag"]:
                    h["created_at"] = h["created_at"].strftime("%m/%d/%Y, %H:%M:%S")
                    if "_id" in h:
                        del h["_id"]

                for m in tweet["relations"]["mention"]:
                    m["created_at"] = m["created_at"].strftime("%m/%d/%Y, %H:%M:%S")
                    if "_id" in m:
                        del m["_id"]

                for r in tweet["relations"]["retweet"]:
                    r["created_at"] = r["created_at"].strftime("%m/%d/%Y, %H:%M:%S")
                    if "_id" in r:
                        del r["_id"]

            with open(output_file_path, "w") as f:
                json.dump(fetched_tweets, f)

        except tweepy.error.TweepError as e:
            logger.debug(f"Encountered missing user {uid}.")
            missing_uids.append(uid)

    return missing_uids


def retrieve_user_histories(
    file=None,
    uids=None,
    issue="history",
    since="2020-08-01",
    include_users=False,
    include_relations=False,
    output_folder="./",
):
    os.makedirs(output_folder, exist_ok=True)
    db_conn = SNmongo(
        issue=issue, include_relations=include_relations, include_users=include_users
    )

    twitter_api = credentials.get_api(db_conn, purpose="lookup")[0]

    uids = loader._load_from_file(file) if not uids and file else uids

    missing_uids = _get_batch_histories(
        uids,
        twitter_api,
        db_conn=db_conn,
        issue=issue,
        include_relations=include_relations,
        include_users=include_users,
        output_folder=output_folder,
    )

    with open(os.path.join(output_folder, "missing_uids.json"), "w") as f:
        json.dump(missing_uids, f)


def retrieve_users_history_from_file(
    file,
    issue="history",
    since="2020-08-01",
    api_count=1,
    batch_size=2048,
    include_users=False,
    include_relations=False,
    output_folder="./",
    output_file="missing_users.json",
    parallel=False,
):
    output_file_path = os.path.join(output_folder, output_file)

    uids = loader._load_from_file(file)
    logger.info("Loaded {} uids from file.".format(len(uids)))

    if parallel:
        db_conn = db_client.DBClient(
            issue=issue,
            include_relations=include_relations,
            include_users=include_users,
        )

        uid_batches = loader._get_batches(uids, batch_size=batch_size)

        logger.info(f"Retrieving {api_count} twitter api accesses.")

        twitter_apis = credentials.get_api(db_conn, api_count=api_count)
        max_workers = api_count

        process_batches = activity._get_process_batches(uid_batches, twitter_apis)

        logger.info(
            f"Fetching history for {len(uids)} uids in {len(uid_batches)} batches with {max_workers} processes."
        )
        ts = time.time()
        processing_result = parallel.run_parallel(
            **{
                "fn": _get_batch_histories,
                "kwargs_list": [
                    {
                        "uid_batch": batch["uid_batch"],
                        "twitter_api": batch["twitter_api"],
                        "since": since,
                        "issue": issue,
                        "include_users": include_users,
                        "include_relations": include_relations,
                    }
                    for batch in process_batches
                ],
                "scheduler": "processes",
                "max_workers": max_workers,
                "progress": True,
                "log_done": True,
            }
        )

        missing_users = activity._get_missing_users(processing_result=processing_result)

        logger.info(
            "Found {} missing users from {} in {:.2f}.".format(
                len(missing_users), file, time.time() - ts
            )
        )
        logger.info(f"Saving missing user ids to {output_file}.")
        loader._save_to_json(content=missing_users, output_file=output_file_path)
    else:
        retrieve_user_histories(
            uids=uids,
            issue=issue,
            include_users=include_users,
            include_relations=include_relations,
            output_folder=output_folder,
        )


if __name__ == "__main__":

    parser = argparse.ArgumentParser(description="Load tweets history for users.")
    parser.add_argument(
        "--file",
        type=str,
        help="File containing user ids to fetch history for.",
        required=True,
    )
    parser.add_argument(
        "--output-folder", type=str, help="Output file.", required=False, default="./"
    )
    parser.add_argument(
        "--issue",
        type=str,
        help="Name of db storing collected information.",
        required=False,
        default="history",
    )
    parser.add_argument(
        "--since",
        type=str,
        help="Specify since when should fetch history.",
        required=False,
        default="2020-08-01",
    )
    parser.add_argument(
        "--api-count",
        type=int,
        help="Number of APIs to use in the process.",
        required=False,
        default=1,
    )
    parser.add_argument(
        "--batch-size",
        type=int,
        help="Size of active uids check batches.",
        required=False,
        default=2048,
    )
    parser.add_argument(
        "--include-relations",
        type=str,
        help="include relations from streamed content.",
        required=False,
        default="yes",
    )
    parser.add_argument(
        "--include-users",
        type=str,
        help="include user profiles from streamed content.",
        required=False,
        default="yes",
    )
    parser.add_argument(
        "--parallel",
        type=str,
        help="include user profiles from streamed content.",
        required=False,
        default="no",
    )
    parser.add_argument(
        "--reset",
        type=str,
        help="Reset twitter credentials use settings.",
        required=False,
        default="no",
    )

    args = parser.parse_args()

    file = args.file
    since = args.since
    api_count = args.api_count
    issue = args.issue
    batch_size = args.batch_size
    include_relations = args.include_relations != "no"
    include_users = args.include_users != "no"
    parallel = args.parallel != "no"
    output_folder = args.output_folder
    reset = args.reset != "no"

    if reset:
        db_conn = SNmongo()
        db_conn.reset_twitter_credentials_statuses()

    retrieve_users_history_from_file(
        file=file,
        issue=issue,
        since=since,
        api_count=api_count,
        batch_size=batch_size,
        include_relations=include_relations,
        include_users=include_users,
        parallel=parallel,
        output_folder=output_folder,
    )
